// bagging.cc

#include<strstream.h>
#include"bagging.h"
#include"utils.h"

Bagging::Bagging() : confusion(), nclass(0), 
                     baselist(), modellist(), base_learner_train_errors(), base_learner_test_errors(), ensemble_errors(),
										 maxit(MAX_ITERATIONS), current_it(0)
{ base = NULL; trainset = NULL; testset = NULL;}
										 

/**
	 Constructor.
	 @param baselearner base learner to be used by the bagging algorithm
	 @param nclass number of classes;
*/
Bagging::Bagging(BaseLearner* baselearner, unsigned num_class = DEFAULT_NUM_CLASS) :
  confusion(num_class,num_class,0.0), nclass(num_class), baselist(), modellist(), 
	base_learner_train_errors(), base_learner_test_errors(), ensemble_errors(), maxit(MAX_ITERATIONS), current_it(0)
{
  base = baselearner;
  trainset = NULL; 
	testset = NULL;
}


/**
	 Copy Constructor.
*/
Bagging::Bagging(const Bagging& Bag) {
 // n_attr = Bag.n_attr;
 nclass = Bag.nclass;
 confusion = Bag.confusion;
 
 // copy of the base learners of Bag into this
 baselist=Bag.baselist;
 deque<BaseLearner*>::const_iterator it;
 for (it = Bag.baselist.begin(); (it != Bag.baselist.end()); it++) {
	 baselist.push_back((*it)->clone());
 }  
 
 modellist = Bag.modellist;
 base_learner_train_errors = Bag.base_learner_train_errors;
 base_learner_test_errors = Bag.base_learner_test_errors;
 ensemble_errors = Bag.ensemble_errors;
 maxit = Bag.maxit;
 current_it = Bag.current_it;
 base = Bag.base;
 trainset = Bag.trainset;
 testset = Bag.testset; 
}



/**
	 Destructor.
*/
Bagging::~Bagging() {
  deque<BaseLearner*>::iterator it;
  for (it = baselist.begin(); (it != baselist.end()); it++) 
	  delete *it;
		
	if (trainset != NULL) {
	  delete trainset;
		trainset = NULL;
  }
	
	if (testset != NULL) {
	  delete testset;
		testset = NULL;
  }

}


/*
	 Trains the bagged ensemble using a set of resampled data sets.\\
*/
unsigned Bagging::train_base_model (char* basename, char* basemodel, unsigned startiteration,
                                   unsigned maxiterations, bool save_err) {
  
	char trainfile [256];
	char modelfile [256];
  int it;
  float err;
	
  ostrstream otrain(trainfile,256,ios::out);
	ostrstream omodel(modelfile,256,ios::out);

  for (it = startiteration; it <= maxiterations; it++) {
	   // building name of the trainig files and of the relative models
     otrain.seekp(0,ios::beg);
		 omodel.seekp(0,ios::beg);		 
		 otrain << basename << it << ends;
		 omodel << basemodel << "-m" << it << ends;
		 // cloning and training of the base learner
	   BaseLearner* b = base->clone();
	   b->set_number(++current_it); 
	   err = b->train_learner(trainfile,modelfile);
		 // putting base learner, error and name of the file model into lists
     baselist.push_back(b);
	   base_learner_train_errors.push_back(err);
		 string s(modelfile);
		 modellist.push_back(s);
  }
  if (save_err) {
	  omodel.seekp(0,ios::beg);
		omodel << basemodel << SUFFIX_BASE_LEARNER_ERRORS << ends;
		save_base_learner_errors(modelfile, base_learner_train_errors);
	}
  return (current_it);
}




/*
	 Combines the bagged ensemble of trained base learners using model files.
*/
float Bagging::combine_models (char* testfile, char* predicted, unsigned startiteration,
                              unsigned maxiterations, bool save_err) {

  int predicted_class;
	int trueclass;
	int num_errors = 0;
	unsigned npattern = 0;
	float err;
	unsigned const SIZE = 256;
	char * const BASRES = "basres";
	char model[SIZE];
	char results[SIZE];
	char testlabels[SIZE];
	unsigned * vclasses = new unsigned[nclass];
	unsigned k;
	
	ifstream * ifs = new ifstream[maxiterations-startiteration+1];
	
	// testing of the base learners on the file testfile. Predicted classes are stored in files basres	
	ostrstream omodel(model,SIZE,ios::out);
	ostrstream oresults(results,SIZE,ios::out);
	for (int i=startiteration; i<=maxiterations; i++) {	   
		 omodel.seekp(0,ios::beg);	
		 oresults.seekp(0,ios::beg);
		 if ((i>modellist.size()) || (i<1))
		   error_exit("Bagging::combine_models : modellist out of range");	 		 
		 omodel << modellist[i-1] << ends;
		 oresults << BASRES << i << ends;
		 if ((i>baselist.size()) || (i<1))
		   error_exit("Bagging::combine_models : baselist out of range");	
	   err = (baselist[i-1])->test_learner(testfile, model, results);
		 base_learner_test_errors.push_back(err);
	}
	
	// Combining the results using majority voting and the files basres
	// opening test labels file
	ostrstream otestlabels(testlabels,SIZE,ios::out);
	otestlabels << testfile << ".labels" << ends;
	ifstream ftest(testlabels);	
	if (ftest == NULL)
	  error_exit("Bagging::combine_models : Can't open file ", testlabels);
	// opening file of the classes predicted by the ensemble
	fstream fpredicted(predicted,ios::in | ios::out);	
	if (fpredicted == NULL)
	  error_exit("Bagging::combine_models : Can't open file ", predicted);
	// opening files of the predicted classes computed by the base learners
	k = 0; // recall that startiteration can be not equal to 0
	for (int i=startiteration; i<=maxiterations; i++) {
	  oresults.seekp(0,ios::beg);
	  oresults << BASRES << i << ends;
	  ifs[k++].open(results);
		if (ifs[k-1] == NULL)
		  error_exit("Bagging::combine_models : Can't open file ", results);
	}
	
	// Reading the classes computed by the base learners, one pattern at a time and storing
	// them into the vector vclasses and computes the predicted class by majority voting
	npattern = wc(testlabels);
	unsigned pclass;
	for (int j=0; j<npattern; j++) {
	  for (int t = 0; t<nclass; t++)
		  vclasses[t] = 0;
		k = 0; // recall that startiteration can be not equal to 0
		for (int i=startiteration; i<=maxiterations; i++) {
		  ifs[k++] >> pclass;
			vclasses[--pclass]++;
		}
		predicted_class = majority_voting(vclasses);
		fpredicted << predicted_class << endl;
	}
	
	// closing files of the predicted classes computed by the base learners
	k = 0; // recall that startiteration can be not equal to 0
	for (int i=startiteration; i<=maxiterations; i++) {
	    if (ifs[k].bad())
			  error_exit("Bagging::combine_models : Error reading classes predicted by the base learners ");
		  ifs[k++].close();
	}
	
	// Computing the error rate of the ensemble and the confusion matrix
	unsigned num_err = 0;
	fpredicted.seekg(0,ios::beg);	 
	ftest.seekg(0,ios::beg);
	confusion.initialize();	
	for (unsigned i = 0 ; i<npattern; i++) {
	  ftest >> trueclass;
		fpredicted >> predicted_class;
		confusion(predicted_class-1, trueclass-1) += 1;
		if (predicted_class != trueclass)
		  num_err++;
	}
	
	err = (float)num_err/(float)npattern;
	if (save_err) {
	  omodel.seekp(0,ios::beg);
		omodel << predicted << SUFFIX_BASE_LEARNER_ERRORS << ends;
		save_base_learner_errors(model, base_learner_test_errors);
	}		
	
	// cleaning up
	delete [] vclasses;
	delete[] ifs;
	
	return err;
}	
	
	
void Bagging::save_base_learner_errors(char* filename, deque<float>& b) {
  ofstream of(filename);
	if (!of)
	  error_exit("Bagging::save_base_learner_errors : Can't open file ", filename);
  for (unsigned i = 0; i < b.size(); i++)
	  of << b[i] << endl;
	if (of.bad())
	  error_exit("Bagging::save_base_learner_errors : Error writing file ", filename);
}

unsigned Bagging::majority_voting(unsigned* vclasses) {
  unsigned predicted;
	unsigned maxx = 0;
	deque<unsigned>maximal_classes;
	
	for (int i=0; i<nclass; i++) {
	  if (vclasses[i] > maxx) {
		  maxx = vclasses[i];
			maximal_classes.clear();
			maximal_classes.push_back(i+1);
		} else if (vclasses[i] == maxx)
		  maximal_classes.push_back(i+1);
	}
	if (maximal_classes.size() == 1)
		  predicted = maximal_classes[0];
	else {
		  long seed = (long) -time(NULL);
			unsigned guess = (unsigned)(ran1(&seed)*maximal_classes.size());
			predicted = maximal_classes[guess];
	}
	return predicted;		
}

/*
	Prints the confusion matrix to an output stream
*/ 
void Bagging::print_conf_matrix(ostream& os = cout){
  
	os << "Confusion matrix of the bagged ensemble:" << endl;
	os.precision(0);
  confusion.print(os);
}
