/* ******************************************************************************
NEURObjects 2.01
bagging.h January 2002
Modified  May 2002
Giorgio Valentini
****************************************************************************** */
#ifndef BAGGING_H
#define BAGGING_H

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include "matrix.h"
#include "tset.h"
#include "baselearner.h"

using namespace std;

static const char * const SUFFIX_BASE_LEARNER_ERRORS = ".blerr";

static const char * const SUFFIX_ENSEMBLE_ERRORS = ".ens_errors";

// base name of the result files computed by the base learners
static const char * const BASRES = "basres";

/**
   Class building  ensembles of bagged base learners.\\
	 This class implements the general Breiman's learning algorithm of bagging, selecting a class by majority voting
	 among the trained bagged base learners.
	 The base learners must belong to a derived class of BaseLearner. A pointer to a specific base learner (base) is passed to the
	 constructor: each new added base learner will be a copy  of base. 
	 Learned models and bootstrapped data sets are stored in files. Bootstrapped data sets are stored in files consecutively
	 numbered (the number are the last character of the common name of the data sets (that is name1, name2, ... , namen).
	 This class permits to easily merge different bagged ensemble, and it allows to use different base learners 
	 (as long as the base learners are derived classes of the abstract virtual base class Baselearner) 
*/ 
class Bagging {

protected:
/// Base learner model. Each new added base learner is equal to base.
BaseLearner* base;

/// Default starting number of iterations
static const unsigned START_ITERATIONS = 1;

/// Default maximun number of iterations
static const unsigned MAX_ITERATIONS = 100;

/// Default  number of classes
static const unsigned DEFAULT_NUM_CLASS = 2;

/// Default  number of attributes
static const unsigned DEFAULT_NUM_ATTR = 2;


/// Confusion matrix
matrix confusion;

/// number of classes
unsigned nclass;


/**
	 Computes the files of the classes predicted by the ensemble.\\
	 A test file named testfile.labels must be provided. It contains the list of the labels
	 of the classes of the test file in the same format of the predicted file of the class BaseLearner.
	 At each iteration it stores  error test rate of each base learner in the list base_learner_test_errors.
	 The results (list of the predicted classes)  must be stored  in  files from 
	 basres1 to basresn, where n id the number of the base learners,
	 Then all these files are opened and each pattern is read one at a time from all the files and 
	 the predicted class of the enesemble is assigned through majority voting and stored in the file predicted.
	 @param testfile  name of the testing set.
	 @param predicted file storing the predicted classes. 
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 	
*/
virtual void compute_ensemble_results(char* testfile, char* predicted, unsigned startiteration, unsigned maxiterations);

/**
	 Computes the ensemble errors at each iteration.\\
	 Computes the ensemble errors at each iteration and stores in the file predicted the classes predicted by the ensemble at the last
	 iteration.
	 A  file named testfile.labels must be provided. It contains the list of the labels
	 of the classes of the test file in the same format of the predicted file of the class BaseLearner.
	 The results computed by the base learners (list of the predicted classes)  must be stored  in  files from 
	 basres1 to basresn, where n id the number of the base learners.
	 Then these files are opened one at a time from  and the predictions are stored in a matrix (npattern*nclass).
	 At each iteration the predicted class of the ensemble is assigned through majority voting. The errors of the ensemble at each
	 iteration is stored in the list ensemble_errors and in the file predicted.ens_errors. If the flag
	 save_pred_at_each_iteration is set, at each iteration the predictions of the ensemble are saved in the files
	 predicted-startiteration .. predicted-maxiterations.
	 The  classes predicted by  the ensemble at the last iteration are stored in the file predicted.
	 @param testfile  name of the testing set.
	 @param predicted file storing the predicted classes. 
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 	
	 @param save_pred_at_each_iteration if true the predicted classes are saved onto a file at each iteration
*/
virtual void compute_ensemble_results_step(char* testfile, char* predicted, unsigned startiteration, unsigned maxiterations,
                                           bool save_pred_at_each_iteration = false);

/**
	 Save onto a file the label of the prdicted classes.
	 @param predicted file storing the predicted classes. 
	 @param vpredicted vector of the label predicted by the ensemble
	 @param  npattern number of the patterns to be saved
*/
void save_predicted(char* predicted, const unsigned * vpredicted, unsigned const npattern);

/**
	 Computes the error and the confusion matrix of the ensemble.\\
	 A test file named testfile.labels must be provided. It contains the list of the labels
	 of the classes of the test file in the same format of the predicted file of the class BaseLearner.
	 It opens the file testfile.labels and the file of the classes predicted by the ensemble (file predicted)
	 and reading from them computes the error rate and the confusion matrix.
	 @param testfile  name of the testing set.
	 @param predicted file storing the predicted classes. 
	 @param  save_err if true the errors of the base learners are stored into the file predicted.blerr
*/
float calc_error_and_confusion(char* testfile, char* predicted, bool save_err);

public:


// members

/// List of the base learners
deque<BaseLearner*> baselist;

/**  List of the model files.\\
     It stores the names of the files of the trained base learners.
*/
deque<string> modellist;

/**  List of the training errors of the base learners.\\
     It stores the training error rate of each base learner composing the ensemble.
*/
deque<float> base_learner_train_errors;

/**  List of the test errors of the base learners.\\
     It stores the test error rate of each base learner composing the ensemble.
*/
deque<float> base_learner_test_errors;

/**  List of the error rates of the ensemble at each iteration .\\
     It stores the error rate of the ensemble at each iteration.
*/
deque<float> ensemble_errors;

/// Maximun number of iterations(base learners).
unsigned maxit;

/**  Current number of iterations(base learners). \\
The numbering starts from 1, and  the number of the base learners is  current_it;
*/
unsigned current_it;

/// Training set
TrainingSet* trainset;

/// Test set
TrainingSet* testset;

// methods ///////////////////////////

/**
	 Dummy constructor.	
	 It does nothing 
*/
  Bagging();

/**
	 Constructor.
	 @param baselearner base learner to be used by the bagging algorithm
	 @param num_class number of classes;
*/
Bagging(BaseLearner* baselearner, unsigned num_class = DEFAULT_NUM_CLASS);

/**
	 Copy Constructor.
*/
  Bagging(const Bagging& Bag);

/**
	 Assignment operator.
*/
  Bagging& operator=(const Bagging& Bag);

/**
	 Destructor.
*/
 virtual ~Bagging();

/**
	 Returns a specific base learner from the bagged ensembles
	 @param i  index of the base learner
	 @return the base ith base learner
*/
BaseLearner* operator[](int i);

/**
	 Aggregates enesembles of base learners.\\
	 It aggregates two bagged ensembles, returning a new bagged ensemble composed through the fusion of two bagged ensembles
*/ 
Bagging operator+(const Bagging& Bag);

/**
	 Overloading += operator.\\
	 It aggregates two bagged ensembles, adding the base learners and the modellist of the bagged ensemble on the right to
	 the bagged ensemble on the left.
*/
Bagging& Bagging::operator+=(const Bagging& Bag);

/**
	Prints info about the bagged ensemble
	@param O output stream
*/ 
void print_info(ostream& O);

/**
	Prints info about the bagged ensemble to cout
*/ 
void print_info() { print_info(cout); }

/**
	Prints the confusion matrix to an output stream
*/ 
void print_conf_matrix(ostream& os = cout);

/**
	Gets the confusion matrix 
*/ 
inline matrix & get_conf_matrix(void){return confusion;}

/**
	Gets the pointer to the base learner of the ensemble
	@return pointer to a base learner
*/ 
BaseLearner* get_base_learner(void){ return base;}

/**
	Sets the base learner of the ensemble
	@param b base learner
*/ 
void set_base_learner(BaseLearner* b){ base = b;}

/**
	Saves into a file the error rate of each base learner of the ensemble
	@param filename file of the error rates of the base learners
	@param b list of the errors
*/ 
void save_base_learner_errors(char* filename, deque<float>& b);

/**
	Saves into a file filename.ens_errors the error rate of the ensemble at each iteration
	@param filename file of the error rates of the ensemble
*/ 
void save_ensemble_errors(char* filename);											 

/**
	Computes by majority voting the class
	Using a vector of the classes voted by the base learners it calculates the class predicted by the ensemble by majority voting.
	If two or more classes achieve the same maximal number of votes a random guess is performed.
	@param vclasses vector of the classes voted by the base learners
*/ 
unsigned majority_voting(unsigned* vclasses);

/**
	 Trains the bagged ensemble using a set of resampled data sets.\\
	 It trains the base learners using data sets previously resampled by bootstrap.
	 The data sets must be named as basenamen, where n varies from startiteration to maxiterations.
	 At each iteration it stores  error training rate of each base learner in the list base_learner_train_errors. 
	 The trained models can be stored in files; the name of the files are stored in the list modellist.
	 The modelllist contain the name of the models (whose name is basemodeln, with n going from startiteration to
	 maxiterations). The models are stored through the method train_learner of the base learner.\\
	 Note that you can call more times this method, eventually using different base learners (calling set_base_learner).
	 @param basename  base name of the training set. 
	 @param basemodel  base name of the models.
	 @param startiteration start number of iterations(base learner). 
	 @param maxiterations maximun number of iterations(base learners).
	 @param  save_err if true the errors of the base learners are stored into the file basemodel.blerr
	 @param b_printout if true print to stdout a point at each iteration
	 @return the number of trained base learners
*/
virtual unsigned train_base_model (char* basename, char* basemodel, unsigned startiteration = START_ITERATIONS,
                                   unsigned maxiterations = MAX_ITERATIONS, bool save_err = false, bool b_printout = true);

/**
	 Test the trained base learners stored in files.\\
	 Testing of the base learners on the file testfile, using models whose names are stored in the list
	 modellist. The format of the testfile is free, but it must be understandable to the base learner.
	 Predicted classes are stored in files basresn, where n range from startiteration to maxiterations.
	 @param testfile  name of the testing set. 
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 
	 @param dir_res	directory where the computed classes are stored
*/
virtual void test_base_learners(char* testfile, unsigned startiteration, 
                       unsigned maxiterations, char* dir_res="");

/**
	 Combines the bagged ensemble of trained base learners using model files.\\
	 It performs a test of the bagged ensembled on an assignated test set (testfile). 
	 The format of the testfile is free, but it must be understandable to the base learner.
	 A second test file named testfile.labels must be provided. It contains the list of the labels
	 of the classes of the test file in the same format of the predicted file of the class BaseLearner.
	 At each iteration it stores  error test rate of each base learner in the list base_learner_test_errors.
	 The names of the files of the trained base learners are stored in the list modellist.
	 Each trained base learner is loaded from the files stored in model list and it is tested on all 
	 the test set. 
	 The results computed by the base learners (list of the predicted classes)  are stored  in  files from 
	 basres1 to basresn, where n id the number of the base learners.
	 Then all the files are opened and each pattern is read one at a time from all the files and 
	 the predicted class of the enesemble is assigned through majority voting.\\
	 Output: error rate, confusion matrix and file of the classes predicted by the ensemble.
	 @param testfile  name of the testing set. 
	 @param predicted file storing the predicted classes.
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 
	 @param  save_err if true the errors of the base learners are stored into the file predicted.blerr
	 @param b_printout if true print to stdout a point at each iteration
	 @return error rate
	 @see BaseLearner
*/
virtual float combine_models (char* testfile, char* predicted, unsigned startiteration = START_ITERATIONS,
                              unsigned maxiterations = MAX_ITERATIONS, bool save_err = false, bool b_printout = true);
															
/**
	 Combines the bagged ensemble of trained base learners using model files storing the results at each iteration.\\
	 It performs a test of the bagged ensembled on an assigned test set (testfile). 
	 The format of the testfile is free, but it must be understandable to the base learner.
	 A second test file named testfile.labels must be provided. It contains the list of the labels
	 of the classes of the test file in the same format of the predicted file of the class BaseLearner.
	 At each iteration it stores  error test rate of each base learner in the list base_learner_test_errors, and
	 the ensemble error in the list ensemble_errors.
	 Each trained base learner is loaded from the files stored in modellist and it is tested on all 
	 the test set. 
	 The results computed by the base learners (list of the predicted classes)  are stored  in  files from 
	 basres1 to basresn, where n id the number of the base learners.
	 Then these files are opened one at a time from  and the predictions are stored in a matrix (npattern*nclass).
	 At each iteration the predicted class of the ensemble is assigned through majority voting. The errors of the ensemble at each
	 iteration is stored in the file predicted.ens_errors.
	 If the flag save_pred_at_each_iteration is set, at each iteration the predictions of the ensemble are saved in the files
	 predicted-startiteration .. predicted-maxiterations. \\
	 Output: error rate, confusion matrix, file of the classes predicted by the ensemble, ensemble error at each iteration.
	 @param testfile  name of the testing set. 
	 @param predicted file storing the predicted classes.
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 
	 @param  save_err if true the errors of the base learners are stored into the file predicted.blerr
	 @param b_printout if true print to stdout a point at each iteration
	 @param b_save_pred_at_each_iteration if true the predicted classes are saved onto a file at each iteration
	 @return error rate
	 @see BaseLearner
*/
virtual float combine_models_step (char* testfile, char* predicted, unsigned startiteration = START_ITERATIONS,
                              unsigned maxiterations = MAX_ITERATIONS, bool save_err = false, bool b_printout = true,
															bool b_save_pred_at_each_iteration = false);

};





#endif
