/* ******************************************************************************
NEURObjects 2.0
basesvml.h January 2002
modified October 2004
Giorgio Valentini
****************************************************************************** */
#ifndef BASESVML_H
#define BASESVML_H

#include <iostream>
#include <fstream>
#include <string>
#include "baselearner.h"

using namespace std;


// suffix of the files containing continuous outputs computed by the SVM
static const char*  OUT = ".out";

//////////////////////////////////////////////////////////////////////////////////////////////////


/**
   Abstract base class for SVM-light base learners.\\
	 The learning is performed reading a training file in SVM-light format and storing the corresponding
	 learned model. The classification is stored in a file stroing the classes according to the BaseLearner
	 class format. Moreover a file with suffix .out containing the computed coninuous outputs of the SVM is
	 generated.
	 
*/ 
class BaseSVM_light_common : public BaseLearner {

protected:

const char* type;

/// trade-off between training error and margin
float C;

/// differential cost error between positive and negative examples
float cost;

/** Read the training error from the stdout of the SVM.\\
    Computes the training error reading the number of amples and the number of misclassification from
    the output of the SVM-light application svm_learn. 
		@param outfile output file 
		@return training error rate
*/
float read_train_error(char* outfile);


/** Read the testing error from the stdout of the SVM.\\
    Computes the testing error reading the number of amples and the number of misclassification from
    the output of the SVM-light application svm_classify.
	  @param outfile output file 
	  @return training error rate
*/
float read_test_error(char* outfile);

/** Saves the predicted classes.\\
    Saves the predicted classes reading the output file of the   SVM-light application svm_classify.
		@param predicted output file  of the predicted classes
*/
void save_predicted(char* predicted); 


public:

/// Get the C parameter
float const get_C() const {return C;}

/// Set the C parameter
void set_C(float val) {C=val;}

/// Get the cost parameter
float const get_Cost() const {return cost;}

/// Set the cost parameter
void set_Cost(float val) {cost=val;}

/**
	 Tests the base learner using a testing set stored in a file.\\
	 The learner is tested using a testing set stored in a file and a  previously learned model  stored in a
	 file, both in SVM-light format.
	 It outputs two files: the predicted classes file in the format specified in the base class BaseLearner,
	 and a file with the same and an added suffix .out that stores the continuous outputs of the SVM.
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by hte ensemble classes to
	 combine the outputs of the base learners.	 
	 @param testfile  name of the testing set. 
	 @param model  model file name.
	 @param predicted computed class file.
	 @return error rate
*/
virtual float test_learner(char* testfile, char* model, char* predicted);

};



//////////////////////////////////////////////////////////////////////////////////////////////////

/**
   Wrapper classes for SVM-light linear SVM.\\
	 
*/ 
class BaseSVM_light_linear : public BaseSVM_light_common {


public:

/// Constructor
BaseSVM_light_linear (float val = 1.0, float c = 1.0)  {num=0; C=val; type = "-t 0 "; cost=c;}


/// Destructor
virtual ~BaseSVM_light_linear () {}

/**
  Method that produces a clone of the current SVM.\\
	@return pointer to the cloned SVM
*/
BaseSVM_light_linear* clone() { return new BaseSVM_light_linear(*this);}

/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. The file formats are in  SVM-light file format.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
virtual float train_learner (char* trainfile, char* model);
											
};




//////////////////////////////////////////////////////////////////////////////////////////////////

/**
   Wrapper classes for SVM-light polynomial SVM.\\
	 The computed kernel funcion is (s a*b +c)^degree.
	 
*/ 
class BaseSVM_light_poly : public BaseSVM_light_common {


// degree of the polynomial kernel
int degree;

// s parameter
float s;

// r parameter
float r;

public:
 
/// Constructor
BaseSVM_light_poly (int valdegree = 2, float valC = 1.0, float s = 1.0, float valr = 1.0, float c = 1.0);

/// Destructor
virtual ~BaseSVM_light_poly () {}

/**
  Method that produces a clone of the current SVM.\\
	@return pointer to the cloned SVM
*/
BaseSVM_light_poly* clone() { return new BaseSVM_light_poly(*this);}

/// Get the degree parameter
int const get_degree() const {return degree;}

/// Set the degree parameter
void set_degree(int val) {degree=val;}

/// Get the s parameter
float const get_s() const {return s;}

/// Set the s parameter
void set_s(float val) {s=val;}

/// Get the r parameter
float const get_r() const {return r;}

/// Set the r parameter
void set_r(float val) {r=val;}


/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. The file formats are in  SVM-light file format.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
virtual float train_learner (char* trainfile, char* model);
										
};




//////////////////////////////////////////////////////////////////////////////////////////////////

/**
   Wrapper classes for SVM-light gaussian SVM.\\
	 The computed kernel funcion is exp(-gamma(a-b)^2), that is gamma is equal to 1/sigma.
	 
*/ 
class BaseSVM_light_gaussian : public BaseSVM_light_common {

// gamma parameter
float gamma;


public:
 
/// Constructor
BaseSVM_light_gaussian (float valgamma = 1.0, float valC = 1.0, float c = 1.0);

/// Destructor
virtual ~BaseSVM_light_gaussian () {}

/**
  Method that produces a clone of the current SVM.\\
	@return pointer to the cloned SVM
*/
BaseSVM_light_gaussian* clone() { return new BaseSVM_light_gaussian(*this);}

/// Get the gamma parameter
float const get_gamma() const {return gamma;}

/// Set the gamma parameter
void set_gamma(int val) {gamma=val;}


/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. The file formats are in  SVM-light file format.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
virtual float train_learner (char* trainfile, char* model);
										
};


//////////////////////////////////////////////////////////////////////////////////////////////////

/**
   Wrapper classes for SVM-light sigmoid SVM.\\
	 The computed kernel funcion is tanh(s a*b +c).
	 
*/ 
class BaseSVM_light_sigmoid : public BaseSVM_light_common {

// s parameter
float s;

// r parameter
float r;

public:
 
/// Constructor
BaseSVM_light_sigmoid (float valC = 1.0, float s = 1.0, float valr = 1.0, float c = 1.0);

/// Destructor
virtual ~BaseSVM_light_sigmoid () {}

/**
  Method that produces a clone of the current SVM.\\
	@return pointer to the cloned SVM
*/
BaseSVM_light_sigmoid* clone() { return new BaseSVM_light_sigmoid(*this);}

/// Get the s parameter
float const get_s() const {return s;}

/// Set the s parameter
void set_s(float val) {s=val;}

/// Get the r parameter
float const get_r() const {return r;}

/// Set the r parameter
void set_r(float val) {r=val;}


/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. The file formats are in  SVM-light file format.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
virtual float train_learner (char* trainfile, char* model);
										
};















#endif
