#ifndef BOOTSTRAPLIGHT_H
#define BOOTSTRAPLIGHT_H

#include<iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cctype>
#include <vector>
#include "matrix.h"

// suffix for the training sets  
  static const char * const TRAIN_B = ".train";
	// suffix for the testing sets  
	static const char * const TEST_B = ".test";

/**
   Class for bootstrapping data in SVM-light format.\\
*/ 
class BootstrapLight {
  
	/// file from which data are randomly drawn
	char* dataset;
	/// base name of the files to be generated
	char* basename; 
	/// number of files to be generated
	int nfiles; 
	/// number of data of the main file dataset
	int ndata;
	/// number of data for each bootstrapped sample
	int ndata_sample;
  /// container of the lines of the data to be drawn
  vector<string> strings;
	/**
	   Matrix storing the results of bootstrapping.\\
		 It has nfiles rows (one for each bootstrapped sample) and ndata columns.
		 If m[b][j] = 0 it means that in the bth sample the jth example of the data set is present.
		 If m[b][j] = 1 it means that in the bth sample the jth example of the data set is not present 
		 (that is this is an out-of-bag point). 
	*/
	short ** m;
	/// seed of the random generator
	int seed;
	
	
	/**
	 Read SVM-light data from a file.\\
	 It loads the vector strings with the lines read from a data set
	 @param dataset   file from which data are read
	 */
	void read_dataset(char* dataset); 
	
	/**
	 Random number generator\\
	 It generates a random integer;
	 @param min minimum integer that will be generated
	 @param max maximum integer that will be generated
	 @return random integer
	 */
	int unif_rand_generate(int min=0, int max=100);
	
	
//////////////////////////////////////
public:

  
				  
	 /// Dummy constructor
   BootstrapLight (void)
	 {}
	 

	 /**
	 Constructor.\\
	 It reads from file the attributes and the class of the patterns and geenrate bootstrap samples.
	 @param data   file from which data are randomly drawn
	 @param base base name of the files to be generated
	 @param nf number of files to be generated
	 @param n number of data per file
	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 */
   BootstrapLight ( char* data, char* base, int nf, int n, int seed = 0) ;
	 
	 /**
	 Constructor.
	 It reads from file the attributes and the class of the patterns.\\
	 It load only the data without generating any file;
	 @param data   data set to be read

	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 */
  BootstrapLight ( char* data, int s = 0);
	 
	 /// Destructor
   virtual ~BootstrapLight () ;
	 
	 /**
	 Extract and store bootstrap samples from the dataset\\
	 It extracts and stores bootstrap samples from the data set acc0rding to the paramenters provided by the
	 constructor
	 @param basename basename of the bootstrapped sample files
	 @param nf number of the bootstrapped files to be generated
	 */
  void write_bootstrap_samples(char* basename, int nf);
	
	
	/// Returns the matrix storing the results of bootstrapping.
	short ** get_m(void) {return m;}
	
	/**
	 Save the matrix storing the results of bootstrapping.\\
	 It has nfiles rows (one for each bootstrapped sample) and ndata columns.
		 If m[b][j] = 0 it means that in the bth sample the jth example of the data set is present.
		 If m[b][j] = 1 it means that in the bth sample the jth example of the data set is not present 
		 (that is this is an out-of-bag point). 
	 @param file name of the file where the matrix will be stored
	 */
	void save_m(char* file);
	
	/**
	 Splits the data set in training and test samples\\
	 It splits the data set basename in multiple training and test set randomly splitted with examples
	 balanced between classes. 
	 The generated files are named basename1.train, basename1.test, ... , bssenamen.train, basenamen.test
	 @param basename name of the  file to be splitted
	 @param n number of the splitted files to be generated
	 @param ratio ratio of the data set reserved for the training data
	 */
	void write_splitted_samples(char* basename, int n, float ratio);
	
	/**
	 Splits the data in different folds\\
	 It splits the data set basename in multiple folds with examples balanced between classes. 
	 The generated data can be used for k-fold cross validation techniques.
	 The generated files are named basename1.f.train, basename1.f.test, ... , bssenamek.f.train, basenamek.f.test
	 @param basename name of the  file to be splitted
	 @param k number of the folds
	 */
	void write_folded_samples(char* basename, int k);

  /**
   Extract and convert class labels fron light data file.\\
	 It read from a light file the class labels, converts them in NEURObjects format and stores them
	 in a file lightdata.labels
	 @param lightdata light file from which the labels are extracted
  */
  void extract_labels_light (char* lightdata);
	
	/// Returns the number of bootstrapped samples
	int get_nfiles(void) {return nfiles;} 
	/// Set the number of bootstrapped samples
	void set_nfiles(int n) {nfiles = n;} 
	/// Returns the number of data of the main file dataset
	int get_ndata(void) {return ndata;}
	/// Set the number of data of the main file dataset
	void set_ndata(int n) {ndata = n;}
	/// Returns the number of data for each bootstrapped sample
	int get_ndata_sample(void) {return ndata_sample;}
	/// Set the number of data for each bootstrapped sample
	int set_ndata_sample(int n) {ndata_sample = n;}
	/// Returns the container of the data
	vector<string> & get_strings(void) {return strings;}                                       
} ;


#endif
