#ifndef BV2DECOMP_H
#define BV2DECOMP_H

#include<iostream>
#include<cstdio>
#include <fstream>
#include <strstream>
#include <algorithm>
#include <string>
#include <cstdlib>
#include <cstring>
#include <deque>
#include "matrix.h"

using namespace std;

struct AverageBV {

  /// Average loss
  float loss;
	/// Average bias
	float bias;
	/// Average net-variance
	float net_variance;
	/// Average variance
	float variance;
	/// Average unbias-var
	float unbias_var;
	/// Average bias-var
	float bias_var;	
	
	/**
	  Constructor.
		@param x intializaton value for all members
	*/
	AverageBV(float x=0);	
	
	/**
	  Initializaes all the members with the same value
		@param x intializaton value for all members
	*/
	void init(float x=0);
	
	/**
	  Print bias-variance values on the stream out
		@param out output stream
	*/
	void print (ostream& out = cout);
};


/**
   Class for evaluating the bias variance decompostion of the error.\\
	 It evaluates the bias variance decomposition of the error using the 0/1 loss for two-class
	 classification problems.
*/ 

class BV2decomp { 
	
  static const int DIMBUF = 256;
	
	/// Number of the datasets (predictions) used to evaluate BV decomposition
	int numdatasets;
	/**
	    Basename of the files containing the predicted outputs.\\
			The files are predbase1.out, predbase2.out, ... predbasenumdatasets.out
	*/
  char predbase[DIMBUF];
	/// testset file
  char datatest[DIMBUF];
  /// Cardinality of the data sets
	int dimtestset;
	
	
	
	/// vector of the prob. for class 1
	float* p1;
	/// vector of the Main predictions
	int* Main;
	/// vector of the labels (targets): +1 class 1, -1 class 2. 
	int* T;
	/// Bias vector. The bias can be 0 or 1
	int* B;
	/// Unbiased variance vector
	float* Vu;
	/// Biased variance vector
	float* Vb;
	/// Net variance vector
	float* NV;
	/// Variance vector
	float* V;
	/// Expected error vector
	float* E;
	/** matrix of the predictions.
	    It collects the predictions made by the base learners on the test set.
	    It is a matrix numdatasets * dimtestset
	*/
	float** P;	
	/// vector of the margins M(x) = (2B(x)-1)(2NV(x)-1)
	float * M;
	/// Vector of the abscissa values of the cumulative distribution of margins
	deque<float> Mvalues;
	/// Vector of the  values of the cumulative distribution of margins
	deque<float> Mcumul;
	
	/**
	  It contains the average values of bias, net-variance, variance, unbiase and biased variance and loss.
	*/ 
	AverageBV AvBV;
	
	/// Matrix of the bootstrap examples
	short **bsm;

/**
	    Counts the number of lines containing data.\\
			It counts the lines of the data. Lines starting with '#' are skipped.			
			@param dataset file containing the data
		
	*/
int count_data_lines(char* dataset);


/**
	    Il loads the matrix of the predictions.
			The predictions performed by the base learners are read from the files
			predbase1, ..., predbasenumdatasets and stored in the matrix P.
			@see P
*/
void load_predictions_matrix(void);
	
///////////////////////////////////////////////////////////////////////
public:

/// Dummy constructor
BV2decomp();
  
/**
	    Constructor.
			It loads the predictions made by the base learner in the matrix P.			
			@param testset test set file name
			@param predictedbasename basename of the files containing the predicted outputs.
			The files are predictedbasename1.out, predictedbasename2.out, ... predictedbasenamendatasets.out.
			@param ndatasets number of the prediction files
*/
BV2decomp(char* testset, char* predictedbasename, int ndatasets);	

/**
	    Destructor.
*/
~BV2decomp(void);

/**
  It computes the probabilities for class 1 
*/
void calc_p1(void);

/**
  It computes the probabilities for class 1 using out-of-bag predictions
	@param m matrix storing the out of bag examples for each predictor. m[i][j] = 1 means that the ith predictions
	on the jth examples is an out of bag prediction.
*/
void calc_p1_oob(short ** m);

/**
  It computes the main predictions by majority
*/
void calc_Main(void);

/**
  It load the target vector.\\
	It reads the labels of the targets from the datatest
*/
void load_T(void);

/// Compute Bias
void calc_Bias(void);

/**
  It computes unbiased, biased and net variance
*/
void calc_Net_Variance(void);

/**
  It computes  variance
*/
void calc_Variance(void);

/**
  Computes Expected error
*/
void calc_Error(void);

/**
  Computes the average values of bias-variance decomposition
*/
void calc_Averages(void);

/**
  Computes the margins
*/
void calc_margins(void);

/**
  Computes the cumulative distribution of the margins
*/
void calc_cumulative_distr_margins(void);

/**
  Save onto oa file the cumulatine distribution of the margins
	@param file file where the cumulative distribution of the margins will be saved
*/
void save_cumulative_distr_margins(char* file);

/**
  It estimates the bias-variance decomposition of the error using the entire test set.\\
	It returns in AvgBV the average values of loss, bias, variance, net, unbiased and biased variance
	@param AvgBV Average values of bias, variance, unbiased and biased variance.
*/
void calc_bv_decomposition(AverageBV& AvgBV);

/**
  It estimates  the out-of-bag bias-variance decomposition of the error.\\
	It uses the out-of-bag examples to estimate the bias variance decompostion of the error.
	It returns in AvgBV the average values of loss, bias, variance, net, unbiased and biased variance
	@param AvgBV Average values of bias, variance, unbiased and biased variance.
	@param m matrix storing the out of bag examples for each predictor. m[i][j] = 1 means that the ith predictions
	on the jth examples is an out of bag prediction. 
*/
void calc_oob_bv_decomposition(AverageBV& AvgBV, short **m);

/**
	  Print bias-variance values on the stream out
		@param out output stream
*/
void print (ostream& out = cout) { AvBV.print(out); }


/**
  Save onto oa file the bias-variance values for each example.\\
	It saves the values of loss, bias, net-variance, unbiased and iased variance and variance for each
	example onto a file.
	Each line corresponds to a different example and the values of loss, bias, net-variance, unbiased
	variance, biased variance and variance are saved (exactly in this order. An header stores some
	information about the file.
	@param file file where the values are stored
*/
void save_bv_per_sample(char* file);

/**
  It loads the bootstrap matrix.\\
	It reads from a file the bootstrap matrix m.
	It has nfiles rows (one for each bootstrapped sample) and ndata columns.
		 If m[b][j] = 0 it means that in the bth sample the jth example of the data set is present.
		 If m[b][j] = 1 it means that in the bth sample the jth example of the data set is not present 
		 (that is this is an out-of-bag point). \\
	Note that the matrix is allocated in the heap by the method. The deallocation is performed by destructor.
	@param file file of the bootstrap matrix
	@return pointer to the generated bootstrap matrix
*/
short ** get_bs_matrix(char* bsmatrix);

                                     
} ;


  

















#endif
