/* ******************************************************************************
NEURObjects 2.02
multipleLobag.h December 2002
Giorgio Valentini
****************************************************************************** */
#ifndef MULTIPLELOBAG_H
#define MULTIPLELOBAG_H

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include <cstdlib>
#include <cstring>
#include "matrix.h"
#include "tset.h"
#include "baselearner.h"
#include "basesvml.h"
#include "bagging.h"
#include "bootstrapLight.h"
#include "bv2decomp.h"
#include "lobag.h"

using namespace std;


// base name of the boostrapped data
static char* const BS_NAME_M = "bs";

// base name of the models
static char* const MODEL_NAME_M = "m";

// base name of prediciton files
static char* const BASEPRED = "prediction";

// default directory for the splitted data
static char* const  DIRSPLIT = "SPLIT";

// default directory for the results computed by the base learners
static char* const DIRRES = "RES";

/**
   Class to perform lobag on multiple data sets.\\
	 It performs the lobag algorithms on different data sets, e.g. performing out of bag estimate of the
	 error (using the class lobag) on multiple training data obtained by multiple hold-out of the data set or
	 by cross-validation. It estimates the error also on the separated test set obtained by hold-out method or
	 cross-validation.
	 
*/ 
class MultipleLobag {

protected:

/// Base learner model. Each new added base learner is equal to base.
BaseSVM_light_common* base;

/// learning set file
char learning_set [256];

/// learning set file without the directories
char basetrain [256];

/// number of the data of the dataset
int ndata;

/// number of bootstrap samples
int B;

/// Dimension of the data
int diminput;


/** 
  Container of the model parameters with their corresponding loss, bias and variance averaged on multiple
	data sets
*/
deque<model_BV> Vavg;

/** 
  Container of the standard deviations of loss, bias and variance whose corresponding average values are
	stored in Vavg
*/
deque<AverageBV> Vsd;

/** 
  Containers of the model parameters with their corresponding loss, bias and variance averaged for each
	separated training set
*/
deque<model_BV> * V;

/// Number of the multiple data (number of splits or number of the folds)
int n;

/// Container of the model parameters to be evaluated
deque<par_model> Param;

/// model of the selected lobag ensemble 
model_BV min_bias;

/// model of the selected bagged ensemble 
model_BV min_loss;

/// minimum oob bias of the evaluated models
float B_min;

/// minimum oob loss of the evaluated models
float L_min;

/// minimum oob bias of the model with the minimum error
float B_L_min;

/** 
  minimum loss of the single base learner.\\
	Estimated loss through multiple hold-out or cross-validation techniques
*/
float loss_single; 
/** 
  minimum loss of the bagged ensemble.\\
	Estimated loss through multiple hold-out or cross-validation techniques
*/
float loss_bag; 
/** 
  minimum loss of the lobag ensemble.\\
	Estimated loss through multiple hold-out or cross-validation techniques
*/
float loss_lobag;

/// Standard deviation of the minimum loss of the single base learner.
float stdev_single;
/// Standard deviation of the minimum loss of the bagged ensemble.
float stdev_bag;
/// Standard deviation of the minimum loss of the lobag ensemble
float stdev_lobag;

/// directory where the splitted and bootstrapped data are stored
char * dirsplit;

/// directory where the predictions and the results files are stored
char * dirres;

/// file storing the parameters of the models
char* modelfile;

/**
	 Print model parameters.
	 It prints the parameter of the models specified in mbv.
	 @param mbv model parameters
	 @param out stream where the parameters are printed
*/
void print_oob_parameters(const model_BV& mbv, ostream& out = cout);

/**
	 Returns the dimension of the input file.\\
	@return dimension of the learning set
*/
int get_diminput(void);

/**
	 Print test errors.
	 It prints the values of the variables loss_lobag, loss_bag, loss_single and the corresponding standard
	 deviations.
	 @param out output stream
*/
void print_test_error(ostream& out = cout);

/////////////////////////////////////////////////////
/////////////////////////////////////////////////////
public:

/// Default constructor
MultipleLobag();

/**
	 Constructor.
	 Model file format:\\
	 Kernel: k\\
	 c1 c2 c3 ... cn\\
	 end\\
	 d1 d2 d3 ... dn opp\\
	 s1 s2 s3 ... sn \\
	 end\\
	 Feasible values for k are 0 (linear), 1(polynomial), 2(gaussian), 3(sigmoid).\\
	 c1 c2 c3 ... cn correspond to the list of the C regularization values.\\
	 If k=0 the lines starting with d1 and with s1 are not present.\\
	 If k=1 is present only the line d1 d2 d3 ... dn corresponding to the polynomial degree.\\
	 If k=2 or k=3 is present only the line  s1 s2 s3 ... corresponding to the sigma values for gaussian or to the
	 multiplicative coefficient for sigmoid kernels.\\	 
	 @param modelfile file storing the information about the models to be evaluated
	 @param data   file from which data are randomly drawn
	 @param nf number of bootstrapped files to be generated
	 @param m number of the multiple data sets (number of splits or folds)
	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 @param dir directory where the splitted or cross-validated and bootstrapped data data will be stored 
	 @param dirp directory where the results will be stored
*/
MultipleLobag (char* modelfile, char* data, int nf, int m=10, int seed = 0, 
               char* dir = DIRSPLIT, char* dirp = DIRRES);


/// Destructor
~MultipleLobag();

             
/** 
   Prints the values of bias-variance decomposition for all the models.
	 It prints the values of the loss, bias, net, unbiased and biased variance and of the pure variance
	 of all the considered models.
	 @param out output stream where the values are printed
*/
virtual void print_BV(ostream& out = cout);

/** 
   Save the values of bias-variance decompostion for all the models.
	 It saves the values of the loss, bias, net, unbiased and biased variance and of the pure variance
	 of all the considered models.
	 @param file file where the values will be saved
*/
void save_BV(char* file);

/** 
   Selects the model for the lobag ensemble.\\
	 It selects from the container Vavg the models with the minimum estimated bias (min_bias), and with the minimum
	 estimated loss (min_loss). It evaluates the minimum oob estimate of the error of the lobag ensemble setting
	 the member B_min that corrsponds to the bias of the model with the minimum oob estimate of the bias,
	 and the oob minimum of the error of the bagged ensemble, setting the member B_L_min that corrsponds to 
	 the bias of the model with the minimum oob estimate of the error. The member B_min is set to the oob estimate
	 of the error of the single SVM.
*/
virtual void Select_model(void);

/** 
   Prints the out-of-bag estimate of the error of lobag and the bagged ensemble.\\
	 It prints the oob estimate of the error of the lobag and bagged ensemble and the estimate of the error of the
	 single SVM.
	 @param out output stream where the values are printed
*/
virtual void print_oob_error (ostream& out = cout);



/**
  Performs multiple exhaustive evaluation of bias-variance decomposition.\\
	It computes the bias-variance decomposition of the error on different training sets, using the class
	Lobag. The results are collected in the vector of containers V.
	In the directory dir must be stored the files of the train data sets. Note that in dir are temporarily stored also
	the models and the predicitons of the SVMs. The oob estimates of the bias-variance decompostion are
	stored in the vector of container V. Each elemant of this vector stores the oob estimates of the
	bias-variance decomposition on a different training set.
	@param data data set
	@param dir directory where the train data are stored
	@param seed seed of the random generator
*/
void multiple_exhaustive_BV_decomp(char* data, char* dir, int seed=0);


/** 
   Computes the average bias-variance between multiple data sets.
	 It averages the bias-variance values stored in the vector of containers V and stores the results in the
	 container Vavg. It stores the corresponding standard deviations in Vsd
*/
void  compute_avg_BV(void);

/**
	 Calculates the loss of the bagged ensemble on mutiple testing data sets.\\
	 It calculates the average loss of the bagged ensemble on multiple testing sets. These test sets can be
	 mutiple hold-out test sets or cross-validation test sets. In any case the bagged ensemble is trained on
	 each training set using the parameters stored in mbv and is tested against a separated test set. The error
	 on the n different test sets is averaged and the standard deviation returned in stdev. The predictions
	 made by the ensemble on each test set are stored in dirres/predicted1, ..., dirres/predictedn.
	 The error at each iteration is stored in dirres/predicted1.ens_errors, ..., dirres/predictedn.ens_errors.
	 @param mbv parameters of the model 
	 @param stdev standard deviation of the error (returned value)
	 @param predicted base file name of the clasees predicted by the ensemble
	 @param seed seed of the random generator.
*/
virtual float Calc_avg_loss(model_BV& mbv, float& stdev, char* predicted, int seed=0);

/**
	 Calculates the loss of the single base learner on multiple testing data sets.\\
	 It calculates the average loss of the single base learner on multiple testing sets. These test sets can be
	 mutiple hold-out test sets or cross-validation test sets. In any case the base learner is trained on
	 each training set using the parameters stored in mbv and is tested against a separated test set. The error
	 on the n different test sets is averaged and the standard deviation returned in stdev. The predictions
	 made by the  base learner on each test set are stored in dirres/predicted1, ..., dirres/predictedn.
	 @param mbv parameters of the model 
	 @param stdev standard deviation of the error (returned value)
	 @param predicted base file name of the clasees predicted by the ensemble
*/
virtual float Calc_avg_loss_base_learner(model_BV& mbv, float& stdev, char* predicted);
          
/** Builds a base learner.\\
It builds a base learner according to the model parameters stored in pm.
    @param pm model parameters
		@return the new base learner
*/
BaseSVM_light_common* build_base_learner (par_model pm);

/** Performs multiple tests.\\
   It performs multiple tests on different test sets, averaging between them. 
	 It estimates the out-of-bag of the error using the set of the training sets, averaging between them.
	 Selects the models for the single SVM, bagged and lobag ensemble and perform the estimate of the their errors
	 on separated (muultiple) test sets, averaging between them. The set of separated trainnig and test sets
	 must be in the directory dirsplit. This method is used both by the methods Multiple_hold_out and
	 Cross_validation. The preditions on different test sets are stored in dirres as
	 learning_set.pred.single[bag][lobag]i, with i ranging from 1 to n.
   @param basepred base name of the generated prediction files
	 @param seed seed of the random generator.
*/
virtual void  Multiple_test  (char* basepred = BASEPRED, int seed=0); 


/** It prints the error of the lobag and bagged ensembles.\\
   It prints the average error and the standard deviation on the test sets.
	 @param out output stream
*/
virtual void print_multiple_test_error(ostream& out = cout);

};


/////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

/**
   Class for performing lobag on multiple data sets using hold out techniques.\\
	 It performs the lobag algorithms on different data sets, e.g. performing out of bag estimate of the
	 error (using the class lobag) on multiple training data obtained by multiple hold-out of the data set.
    It estimates the error also on the separated test set obtained by hold-out methods.
	 
*/ 
class MultipleLobagHoldOut : public MultipleLobag {

public:

/// Default constructor
MultipleLobagHoldOut();

/**
	 Constructor.
	 Model file format:\\
	 Kernel: k\\
	 c1 c2 c3 ... cn\\
	 end\\
	 d1 d2 d3 ... dn opp\\
	 s1 s2 s3 ... sn \\
	 end\\
	 Feasible values for k are 0 (linear), 1(polynomial), 2(gaussian), 3(sigmoid).\\
	 c1 c2 c3 ... cn correspond to the list of the C regularization values.\\
	 If k=0 the lines strating with d1 and with s1 are not present.\\
	 If k=1 is present only the line d1 d2 d3 ... dn corresponding to the polynomial degree.\\
	 If k=2 or k=3 is present only the line  s1 s2 s3 ... corresponding to the sigma values for gaussian or to the
	 multiplicative coefficient for sigmoid kernels.\\	 
	 @param modelfile file storing the information about the models to be evaluated
	 @param data   file from which data are randomly drawn
	 @param nf number of bootstrapped files to be generated
	 @param m number of the multiple data sets (number of splits)
	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 @param dir directory where the splitted or cross-validated and bootstrapped data data will be stored 
	 @param dirp directory where the results will be stored
*/
MultipleLobagHoldOut (char* modelfile, char* data, int nf, int m=10, int seed = 0, 
                      char* dir = DIRSPLIT, char* dirp = DIRRES) : 
                      MultipleLobag(modelfile, data, nf, m, seed, dir, dirp) {};
/**
  It splits the learning set in multiple training and test set.\\
	It randomly splits the data in a separated training and test set multipl times, trying to preserve the
	original proportion of the examples between classes. It calls the methods of the class BootstrapLight
	@param nsplits number of splitted data sets to be generated
	@param ratio ratio of the data set reserved for the training data
	@param dir directory where the splitted data sets will be stored
	@param seed seed of the random generator
*/
void Split(int nsplits, float ratio, char* dir, int seed=0); 

/** It performs a multiple hold-out test on the learning set.\\
   It performs a multiple hold-out test. 
	 It estimates the out-of-bag of the error using the set of the training sets, averaging between them.
	 It performs the out-of-bag selection of the models for the single SVM, bagged and lobag ensemble 
	 and perform the estimate of the their errors on separated (multiple) test sets, averaging between them. 
	 The method generates nsplits random splits of the learning sets. 
	 @param nsplits number of splitted data sets to be generated
	 @param ratio ratio of the data set reserved for the training data
	 @param dir directory where the splitted data sets will be stored
   @param basepred base name of the generated prediction files
	 @param seed seed of the random generator.
*/
void  Multiple_hold_out (int nsplits, float ratio, char* dir, char* basepred = BASEPRED, int seed=0);


/** It prints the error of the lobag and bagged ensembles.\\
   It  prints the multiple hold-out of the error for bagged and lobag ensembles. 
	 It prints the average error and the standard deviation on the test sets.
	 @param out output stream
*/
void print_multiple_test_error(ostream& out = cout);

};
/////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////



/**
   Class for performing lobag on multiple data sets usingcross validation techniques.\\
	 It performs the lobag algorithms on different data sets, e.g. performing out of bag estimate of the
	 error (using the class lobag) on multiple training data obtained by cross-validation of the data set.
    It estimates the error also on the separated test set obtained by cross-validation methods.
	 
*/ 
class MultipleLobagCrossValidation : public MultipleLobag {

public:

/// Default constructor
MultipleLobagCrossValidation();

/**
	 Constructor.
	 Model file format:\\
	 Kernel: k\\
	 c1 c2 c3 ... cn\\
	 end\\
	 d1 d2 d3 ... dn opp\\
	 s1 s2 s3 ... sn \\
	 end\\
	 Feasible values for k are 0 (linear), 1(polynomial), 2(gaussian), 3(sigmoid).\\
	 c1 c2 c3 ... cn correspond to the list of the C regularization values.\\
	 If k=0 the lines strating with d1 and with s1 are not present.\\
	 If k=1 is present only the line d1 d2 d3 ... dn corresponding to the polynomial degree.\\
	 If k=2 or k=3 is present only the line  s1 s2 s3 ... corresponding to the sigma values for gaussian or to the
	 multiplicative coefficient for sigmoid kernels.\\	 
	 @param modelfile file storing the information about the models to be evaluated
	 @param data   file from which data are randomly drawn
	 @param nf number of bootstrapped files to be generated
	 @param m number of the multiple data sets (number of  folds)
	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 @param dir directory where the splitted or cross-validated and bootstrapped data data will be stored 
	 @param dirp directory where the results will be stored
*/
MultipleLobagCrossValidation (char* modelfile, char* data, int nf, int m=10, int seed = 0, 
                      char* dir = DIRSPLIT, char* dirp = DIRRES) : 
                      MultipleLobag(modelfile, data, nf, m, seed, dir, dirp) {};

/**
  Generates folds with the leaning set.\\
	It splits the data set  in multiple folds with examples balanced between classes. 
	The generated data can be used for k-fold cross validation. It calls the methods of the class BootstrapLight
	@param nfolds number of folds
	@param dir directory where the generated data sets will be stored
	@param seed seed of the random generator
*/
void Generate_Folds(int nfolds, char* dir, int seed=0);


/** It performs a cross-validation on the learning set.\\
   It performs a cross-validation of the error of the out-of-bag selelcted lobag and bagged ensemble and of
	 the single base learner. 
	 It estimates the out-of-bag of the error using the set of the training sets, averaging between them.
	 It performs the out-of-bag selection of the models for the single SVM, bagged and lobag ensemble 
	 and perform the estimate of the their errors on the separated folds, averaging between them. 
	 The method generates nfolds folds of the learning sets. 
	 @param nfolds number of folds
	 @param dir directory where the folded data sets will be stored
   @param basepred base name of the generated prediction files
	 @param seed seed of the random generator.
*/
void  Cross_validation  (int nfolds,  char* dir, char* basepred = BASEPRED, int seed=0);

/** It prints the error of the lobag and bagged ensembles.\\
   It  prints the  cross validated estimate of the error for bagged and lobag ensembles. 
	 It prints the average error and the standard deviation on the test sets.
	 @param out output stream
*/
void print_multiple_test_error(ostream& out = cout);

};


/* *******************************************************************************
   MultipleLobagBootstrap
   ****************************************************************************  */ 

/**
   Class for performing lobag on multiple data sets with bootstrap techniques.\\
	 It performs the lobag algorithms on different data sets, estimating the error on a separated
	 test set. The training sets are built by bootstrap from a common learning set.
	 
*/ 
class MultipleLobagBootstrap : public MultipleLobag {

protected:
/// Size of the data botstrapped from the learning set.
int bs_size;

/// test set file
char * testset;

/// test set file without the path of the directories
char basetest [256];

/// Average out-of-bag estimate of the loss of the lobag ensemble
float avg_oob_loss_lobag;

/// Average out-of-bag estimate of the loss of the bagged ensemble
float avg_oob_loss_bag;

/// Average out-of-bag estimate of the loss of the single learning machine
float avg_oob_loss_single;

/// Average out-of-bag estimate of the stdev of the loss of the lobag ensemble
float stdev_avg_oob_loss_lobag;

/// Average out-of-bag estimate of the stdev of the loss of the bagged ensemble
float stdev_avg_oob_loss_bag;

/// Average out-of-bag estimate of the stdev of the loss of the single learning machine
float stdev_avg_oob_loss_single;
  
/** 
  Container of the models with the minimum loss. Each element corresponds to a different bootstrapped training
  set.
*/
deque<model_BV> min_loss_models;

/** 
  Container of the models with the minimum bias. Each element corresponds to a different bootstrapped training
  set.
*/
deque<model_BV> min_bias_models;

/** 
  Errors of the lobag ensembles on the test set
*/
deque<float> errors_lobag;

/** 
  Errors of the bagged ensembles on the test set
*/
deque<float> errors_bag;

/** 
  Errors of thesingle SVMs on the test set
*/
deque<float> errors_single;

public:

/// Default constructor
MultipleLobagBootstrap();

/**
	 Constructor.
	 Model file format:\\
	 Kernel: k\\
	 c1 c2 c3 ... cn\\
	 end\\
	 d1 d2 d3 ... dn opp\\
	 s1 s2 s3 ... sn \\
	 end\\
	 Feasible values for k are 0 (linear), 1(polynomial), 2(gaussian), 3(sigmoid).\\
	 c1 c2 c3 ... cn correspond to the list of the C regularization values.\\
	 If k=0 the lines strating with d1 and with s1 are not present.\\
	 If k=1 is present only the line d1 d2 d3 ... dn corresponding to the polynomial degree.\\
	 If k=2 or k=3 is present only the line  s1 s2 s3 ... corresponding to the sigma values for gaussian or to the
	 multiplicative coefficient for sigmoid kernels.\\	 
	 @param modelfile file storing the information about the models to be evaluated
	 @param data   file from which data are randomly drawn
	 @param test test set file
	 @param nf number of bootstrapped files to be generated
	 @param m number of the multiple data sets (number of data sets bootstrapped from the learning set)
   @pram size size of the data bootstrapped from the learning set
	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 @param dir directory where the splitted or cross-validated and bootstrapped data data will be stored 
	 @param dirp directory where the results will be stored
*/
MultipleLobagBootstrap (char* modelfile, char* data, char* test, int nf, int m=10, int size = 50, int seed = 0, 
                      char* dir = DIRSPLIT, char* dirp = DIRRES);

/**
  Performs multiple tests of ensembles trained on bootstrap samples\\
  It generates bootstrap training samples from the given learning set.
	Then it computes the bias-variance decomposition of the error on different training sets, using the class
	Lobag. The results are collected in the vector of containers V.
	In the directory dir must be stored the files of the train data sets. Note that in dir are temporarily stored also
	the models and the predictions of the SVMs. The oob estimates of the bias-variance decompostion are
	stored in the vector of container V. Each element of this vector stores the oob estimates of the
	bias-variance decomposition on a different training set. The models with the minimum loss and minomum
  estimated oob bias are stored in the containers min_loss_models and min_bias_models (each elements refers to a
  diferrent training set).
   The error on the test set are stored in the containers errors_single, errors_bag, errors_lobag, and the 
   average errors in loss_single, loss_bag and loss_lobag.
  @param dir directory where the bootstrapped training data will be stored.
	@param basepred base name of the generated prediction files
	@param seed seed of the random generator
*/
void  Multiple_bootstrap (char* dir, char* basepred = BASEPRED, int seed=0);

/**
  Generates multiple training samples by bootstrap\\
  It generates multiple bootstrap training samples by bootstrap from the  learning set.
  @param nsample number of the bootstrapped samples
  @param size size of each bootstrapped sample
  @param dir directory where the bootstrapped training data will be stored.
	@param seed seed of the random generator
*/
void  Bootstrap_samples(int nsample, int size, char* dir, int seed=0);

/**
  Performs multiple tests of ensembles trained on bootstrap samples\\
	It computes the bias-variance decomposition of the error on different training sets, using the class
	Lobag. The results are collected in the vector of containers V.
	In the directory dir must be stored the files of the train data sets. Note that in dir are temporarily stored also
	the models and the predictions of the SVMs. The oob estimates of the bias-variance decompostion are
	stored in the vector of container V. Each element of this vector stores the oob estimates of the
	bias-variance decomposition on a different training set. The models with the minimum loss and minomum
  estimated oob bias are stored in the containers min_loss_models and min_bias_models (each elements refers to a
  diferrent training set).
   The error on the test set are stored in the containers errors_single, errors_bag, errors_lobag, and the 
   average errors in loss_single, loss_bag and loss_lobag.
	@param basepred base name of the generated prediction files
	@param seed seed of the random generator
*/
void  Multiple_test  (char* basepred = BASEPRED, int seed=0);

/**
  Performs multiple exhaustive evaluation of bias-variance decomposition.\\
	It computes the bias-variance decomposition of the error on different training sets, using the class
	Lobag. The results are collected in the vector of containers V.
	In the directory dir must be stored the files of the train data sets. Note that in dir are temporarily stored also
	the models and the predicitons of the SVMs. The oob estimates of the bias-variance decompostion are
	stored in the vector of container V. Each elemant of this vector stores the oob estimates of the
	bias-variance decomposition on a different training set.
	@param data data set
	@param dir directory where the train data are stored
	@param seed seed of the random generator
*/
void multiple_exhaustive_BV_decomp(char* data, char* dir, int seed=0);

/** 
   Selects the model for the lobag ensemble.\\
	 It selects multiple models with the minimum estimated bias and with the minimum
	 estimated loss, one for each element of the vector V, storing them in the containers min_loss_models and min_bias_models.  
   Then computes the estimated oob average and stdev of the loss of the lobag, bagged and single SVM.
*/
void Select_model(void);

/** 
   Prints the average out-of-bag estimate of the error of lobag and the bagged ensemble.\\
	 It prints the oob estimate of the error of the lobag and bagged ensemble and the estimate of the error of the
	 single SVM and their standard deviation.
	 @param out output stream where the values are printed
*/
void print_oob_error (ostream& out = cout);

/** 
   Prints the out-of-bag estimate of the error of lobag and the bagged ensemble for all the data sets.\\
	 It prints the oob estimate of the error of the lobag and bagged ensemble and the estimate of the error of the
	 single SVM for all the data sets used.
	 @param out output stream where the values are printed
*/
void print_oob_error_per_dataset(ostream& out = cout);

/** 
   Saves onto a file the out-of-bag estimate of the error of lobag and the bagged ensemble for all the data sets.\\
	 It saves the oob estimate of the error of the lobag and bagged ensemble and the estimate of the error of the
	 single SVM for all the data sets used.
	 @param file name of the file where the errors will be saved
*/
void save_oob_error_per_dataset(char* file);


/**
	 Calculates the loss of multiple bagged ensemble on a single test set .\\
	 It calculates the average loss of the bagged ensemble obtained from differet bootstrapped training set on
   a single test set. The bagged ensemble are trained using the parameters stored in dm and are tested against a separated test set. 
   The predictions made by the ensembles are stored in dirres/predicted1, ..., dirres/predictedn.
	 The error at each iteration is stored in dirres/predicted1.ens_errors, ..., dirres/predictedn.ens_errors.
	 @param dm container of the models to be tested
	 @param stdev standard deviation of the error (returned value)
	 @param predicted base file name of the classes predicted by the ensemble
   @param derr container where the errors of the ensembles trained on different training sets are stored
	 @param seed seed of the random generator.
   @return average loss
*/
float  Calc_avg_loss_bootstrap
 (deque<model_BV>& dm, float& stdev, char* predicted,  deque<float>& derr, int seed=0);
 
/**
	 Calculates the loss of single base learners on a single test set .\\
	 It calculates the average loss of single learners obtained from differet bootstrapped training set on
   a single test set. The single SVMs are trained using the parameters stored in dm and are tested against a separated test set. 
   The predictions made by the single learners are stored in dirres/predicted1, ..., dirres/predictedn.
	 @param dm container of the models to be tested
	 @param stdev standard deviation of the error (returned value)
	 @param predicted base file name of the classes predicted by the ensemble
   @param derr container where the errors of the ensembles trained on different training sets are stored
   @return average loss
*/ 
float Calc_avg_loss_base_learner_bootstrap
  (deque<model_BV>& dm, float& stdev, char* predicted,  deque<float>& derr);

/** It prints the average error of the lobag and bagged ensembles.\\
   It  prints the average error for bagged and lobag ensembles. It prints the error and the standard deviation.
	 @param out output stream
*/
void print_multiple_test_error(ostream& out = cout);

/** It prints the errors on the test set.\\
   It  prints the  error for bagged and lobag ensembles on the test set for each trainnig set used.
	 @param out output stream
*/
void print_multiple_test_error_per_dataset(ostream& out = cout);

/** Saves onto a file the errors on the test set.\\
   It  saves the  error for bagged and lobag ensembles on the test set for each trainnig set used.
	 @param file name of the file
*/
void save_test_error_per_dataset(char* file);

/** 
   Prints the averaged values of bias-variance decompostion for all the models.
	 It prints the values of the loss, bias, net, unbiased and biased variance and of the pure variance
	 of all the considered models, averaging on all the n training sets
	 @param out output stream where the values are printed
*/
void print_BV(ostream& out = cout);

/** Set the current test set.
	 @param test test set file name
*/
void set_testset (char* test) {testset = test;};

};



#endif
