// mutualinfo.h


#ifndef MUTUAL_H
#define MUTUAL_H

#include<iostream.h>
#include<fstream.h>
#include<iomanip.h>
#include<strstream.h>
#include<string.h>
#include"utils.h"



/**
   Class for evaluating mutual information between two random variables.\\
   Random variables are discretized dividing their range values in bins. For each bin is
	 computed its frequency, both for each single variable and the joint frequency.
	 Methods can be used to evaluate the entropy of each variable, the joint entropy,
	 the conditional entropy and the mutual information.
*/ 


class MutualInfo2 {
protected:
  
	/// Default minimum of the random variable value
  static const float MIN_VALUE = 0.0;
  /// Default maximum of the random variable value
  static const float MAX_VALUE = 1.0;	
	
	float min;
	/// Maximum of the random variable value
	float max;
  /// number of data values of each random variable
	unsigned ndata;
  /// number of bins
  unsigned nbins;
	/// Amplitude of a single bin
	float ampl;
	/// Matrix of values of the two random variables. num rows = ndata, num columns = 2
  float **data;
	/// matrix of values of frequencies of random variables; num rows = 2, num columns = nbins 
	float **p;
	/// matrix of joint frequencies between the two variables; num rows = nbins, num columns = nbins 
	float **Pjoint;
	///  Entropy of the two random variables;
	float H[2];
	/// Joint entropy of the two random variables
	float Hjoint;
	/// Mutual information between variables
	float Minfo;
	

  /// Builds data matrix
  void	builddata(unsigned num_data);
	 
  /**
	 Load random variables data from filef to array data. \\
	 First must be less than second
	 @param num_data random variable values cardinality
	 @param f file containing the values of random variables. This file must have columns
	 corresponding to different variables separated by at least a blank.
	 @param first the column in the file of the first variable. The columns are numbered from 1 to up.
	 @param second the column in the file of the second variable. It must be greater than first
	*/ 
  int readdata(unsigned num_data, char *f, unsigned first, unsigned second);
	
	/// It builds and initializes  matrix of single frequecies p
  void buildp(unsigned num_bins);
	
	/// It builds and initializes matrix of joint frequencies
  void buildPjoint(unsigned num_bins);

  /**
	 Calculate the bin of random variable num_var od the  data[num_var][num_data].
	 It calaculates the index of the random variable num_var where the value data[num_var][num_data]
	 falls. num_var is the row of p and the column of data. It corresponds also to the dimension of
	 Pjoint.
	 @param num_var index of the random variable in the matrix data
	 @param num_data index of random variable value in data
	 @return index number of the bin of p
	*/ 
  virtual unsigned assignbin(unsigned num_var, unsigned num_data);
	
	/// Initializes p
	void Init_p (float init=0.0);
	
	/// Initializes Pjoint
	void Init_Pjoint (float init=0.0);
	
public:
   	 
	 /**
	 Constructor.
	 @param num_data random variable values cardinality
	 @param num_bins number of bins by wich each range of values of the 
	        random variables is subdivided
	 @param thedata matrix of data. It must have two columns, each referring to a different random
	        variables. The rows refer to the values of the random variables
	 @param min_v Minimum of the random variable value
	 @param max_v Maximum of the random variable value
	 
	 */
   MutualInfo2 (unsigned num_data, unsigned num_bins,  float **thedata,
	              float min_v = 0.0, float max_v = 1.0);	 
	 /**
	 Constructor.
	 @param num_data random variable values cardinality
	 @param num_bins number of bins by wich each range of values of the 
	        random variables is subdivided
	 @param f file containing the values of random variables. This file must have columns
	 corresponding to different variables separated by at least a blank.
	 @param first the column in the file of the first variable. The columns are numbered from 1 to up.
	 @param second the column in the file of the second variable. It must be greater than first.
	 @param min_v Minimum of the random variable value
	 @param max_v Maximum of the random variable value
	 */
   MutualInfo2 (unsigned num_data, unsigned num_bins,  char *f, 
	              unsigned first, unsigned second, float min_v = 0.0, float max_v = 1.0);
	 
	 /// Destroyer
	 ~MutualInfo2 (); 
	 
	 /// printdata to stdout
	 void printdata(void);
	 
	 /// return minimum value of random variable
	 float Read_min() {return min;}
	 /// return maximum value of random variable
	 float Read_max() {return max;}
	 /// print on stdout minimum value of random variable
	 void Print_min() {cout << "Min. of random variable = " << min << endl;}
	 /// print on stdout maximum value of random variable
	 void Print_max() {cout << "Max. of random variable = " << max << endl;}
	 
	 /**
	 It calculates the frequency of the single random variables.
	 The random variables are discretized in bins and for each bin is calculated the frequency
	 considering the values of the array data. 
	 */
   void Calc_p (void);
	 
	 /**
	 It calculates the joint frequency of the random variables.
	 The random variables are discretized in bins and for each bin is calculated the frequency
	 considering the values of the array data. 
	 */
	 void  Calc_Pjoint (void);
	 
	 /// print the matrix p
	 void Print_p(void);
	 
	 /// print the matrix Pjoint
	 void Print_Pjoint(void);
	 
	 /**
	 It calculates the entropy of the random variables.
	 @return the pointer to the variables entropy. 
	 */
	 float* Calc_Entropy(void);
	 
	 /// Returns the entropy array of the random variables
	 float* Read_Entropy(void){return H;}
	 
	 /// Print to stdout the entropy of the random variables
	 void Print_Entropy(void);
	 	 
	 /**
	 It calculates the joint entropy of the random variables.
	 @return joint entropy of the random variables. 
	 */
	 float Calc_JointEntropy(void);
	 
	 /// Returns the joint entropy  of the random variables
	 float Read_JointEntropy(void){return Hjoint;}
	 
	 /// Print to stdout the joint entropy of the random variables
	 void Print_JointEntropy(void);
	 
	 /**
	 It calculates the mutual information of the random variables.
	 @return mutual information of the random variables. 
	 */
	 float Calc_Minfo(void);
	 
	 /// Returns the mutual information  of the random variables
	 float Read_Minfo(void){return Hjoint;}
	 
	 /// Print to stdout the mutual information of the random variables
	 void Print_Minfo(void);
	 
};
	 

// ============================================================================================

/**
   Class for evaluating the correlation between two random variables representing errors.\\
   Random variables are discretized dividing their range values in bins. For each bin is
	 computed its frequency, both for each single variable and the joint frequency.
	 Methods can be used to evaluate the entropy of each variable, the joint entropy,
	 the conditional entropy and the mutual information. The bins are not necessarily 
	 of the same size. The first bin correspond to correct values, the others to increasing
	 errors. The input data must be errors between the stimated and the "true" values.
	 The Mutual information error is calculated subtracting to the mutual information the mutal
	 information generated by "correct" values or vlues with a single error.
	 @see MutualMatrix
	 @see Outerr
*/ 


class MIError2 : public  MutualInfo2 {

protected:

  /// Mutual information error not considering two or more errors
	float MIErr2;
	
	/// Mutual information generated by correct guess or by a single error
	float MICorrect2;
	
	/// Mutual information error 
	float MIErr;
	
	/// Mutual information generated by correct guess
	float MICorrect;
	
	/** array of the values of the limits of the bins. \\
	    The first value correspond to the leftmost value=min, the last value correspond to the
			rightmost=max. The size of the vector must be nbins + 1.
	*/
	float * bin; 
	
	/** It builds the values of the limits of the bins. \\
	    The first value correspond to the leftmost value=min, the last value correspond to the
			rightmost=max. If binlimits == 0 are builts equispaced bins.
			@param binlimits. Array of the values of the bin limits. The size of the vector must 
			be nbins + 1.
			@see bin
	*/
  void buildbin(float *binlimits = 0);
	
	/**
	 Calculate the bin of random variable num_var od the  data[num_data][num_var].
	 It calculates the index of the random variable num_var where the value data[num_data][num_var]
	 falls. num_var is the row of p and the column of data. It corresponds also to the dimension of
	 Pjoint.
	 @param num_var index of the random variable in the matrix data
	 @param num_data index of random variable value in data
	 @return index number of the bin of p
	*/ 
  virtual unsigned assignbin(unsigned num_var, unsigned num_data);

public:

  /**
	 Constructor.
	 @param num_data random variable values cardinality
	 @param num_bins number of bins by wich each range of values of the 
	        random variables is subdivided
	 @param thedata matrix of data. It must have two columns, each referring to a different random
	        variables. The rows refer to the values of the random variables
	 @param binlimits Array of the limit values of the bins. Its size must be num_bins + 1. If it is
	 not specified the bins are equally spaced between min_v and max_v
	 @param min_v Minimum of the random variable value
	 @param max_v Maximum of the random variable value	 
	 */
   MIError2 (unsigned num_data, unsigned num_bins,  float **thedata, float *binlimits = 0,
	              float min_v = 0.0, float max_v = 1.0);	 
	 /**
	 Constructor.
	 @param num_data random variable values cardinality
	 @param num_bins number of bins by wich each range of values of the 
	        random variables is subdivided
	 @param f file containing the values of random variables. This file must have columns
	 corresponding to different variables separated by at least a blank.
	 @param first the column in the file of the first variable. The columns are numbered from 1 to up.
	 @param second the column in the file of the second variable. It must be greater than first.
	 @param binlimits Array of the limit values of the bins. Its size must be num_bins + 1. If it is
	 not specified the bins are equally spaced between min_v and max_v
	 @param min_v Minimum of the random variable value
	 @param max_v Maximum of the random variable value
	 */
   MIError2 (unsigned num_data, unsigned num_bins,  char *f, unsigned first, 
	           unsigned second, float *binlimits = 0, float min_v = 0.0, float max_v = 1.0);
	 
	 /// Destroyer
	 ~MIError2 (); 
	 
	 /** It calculates the component of mutual information generated by no errors or not correlated
	     errors. \\
			 Considering the matrix of the element of the mutual information between the discretized
			 variables var1 and var2, is computed the mutual info corresponding to the first row and the
			 first column.
	     @return the component of mutual information generated by no errors or not correlated
	     errors
	*/
	float Calc_MICorrect2(void);
	 
	 
	/** It calculates the component of mutual information generated by correlated errors. \\
			 Considering the matrix of the element of the mutual information between the discretized
			 variables var1 and var2, is computed the mutual info corresponding to the overall matrix
			 minus the first row and the first column.
	     @return the component of mutual information generated by correlated errors.
	*/
  float Calc_MIErr2(void); 
	
	
  /** Print to stdout the mutual information MICorrect2. \\
			Print to stdout the mutual information MICorrect2 generated by no errors 
			or not correlated errors
	*/
  void Print_MICorrect2(void);	

  /** Print to stdout the mutual information MIErr2. \\
			Print to stdout the mutual information generated by correlated errors.
	*/
  void Print_MIErr2(void);	
	
	/** It calculates the component of mutual information generated by no errors.\\
			 Considering the matrix of the element of the mutual information between the discretized
			 variables var1 and var2, is computed the mutual info corresponding to the first element
			 of the matrix.
	     @return the component of mutual information generated by no errors.
	*/
	float Calc_MICorrect(void);
	 	 
	/** It calculates the component of mutual information generated by errors. \\
			 Considering the matrix of the element of the mutual information between the discretized
			 variables var1 and var2, is computed the mutual info corresponding to the overall matrix
			 minus the first element
	     @return the component of mutual information generated by errors.
	*/
  float Calc_MIErr(void); 
	
	
  /** Print to stdout the mutual information MICorrect. \\
			Print to stdout the mutual information MICorrect generated by no errors 
	*/
  void Print_MICorrect(void);	

  /** Print to stdout the mutual information MIErr. \\
			Print to stdout the mutual information generated by  errors.
	*/
  void Print_MIErr(void);	

}; // end of class MIError2



// ============================================================================================

/**
   Class for building mutual information among all the couples of a set of random variables.\\
   This class builds matrices of mutual information, joint entropy, and mutual information error 
	 (if the random variables represent errors).	 
	 @see MutualInfo2
	 @see MIError2
	 @see Outerr
*/ 


class MutualMatrix {

private:
  /// size of the file name
	static const unsigned DIMBUF = 128;
  /// Minimum of the random variable value
	float min;
	/// Maximum of the random variable value
	float max;
  /// number of data values of each random variable
	unsigned ndata;
  /// number of bins
  unsigned nbins;
	/// Amplitude of a single bin
	float ampl;
	/// first column of the file to be read
	unsigned int first;
	/// last column of the file to be read
	unsigned int last;
	/** array of the values of the limits of the bins. \\
	    The first value correspond to the leftmost value=min, the last value correspond to the
			rightmost=max. The size of the vector must be nbins + 1.
	*/
	float *bin; 
	/// name of the file to be read
	char namefile[DIMBUF];

  /// Entropy matrix
	float **EnM;
	/// Mutual information matrix
	float **MiM;
	/// Mutual information error matrix not considering single errors
  float **MeM2;
	/// Mutual information error matrix
  float **MeM;
	/// Sum of the element values of the Entropy matrix
	float SumEnM;
	/// Sum of the element values of the Mutual information matrix
	float SumMiM;
	/// Sum of the element values of the Mutual information error matrix not considering single errors
  float SumMeM2;
	/// Sum of the element values of the Mutual information error matrix
  float SumMeM;
	
	/// Number of random variables
	unsigned int nvar; 
	

public:
  
	/**
	Constructor.
		@param num_data random variable values cardinality
		@param num_bins number of bins by wich each range of values of the 
	      	 random variables is subdivided
		@param f file containing the values of random variables. This file must have columns
		corresponding to different variables separated by at least a blank.
		@param first the column in the file of the first variable. The columns are numbered from 1 to up.
		@param last the column in the file of last variable to be considered
		@param binlimits Array of the limit values of the bins. Its size must be num_bins + 1. If it is
		not specified the bins are equally spaced between min_v and max_v
		@param min_v Minimum of the random variable value
		@param max_v Maximum of the random variable value
	*/
  MutualMatrix (unsigned num_data, unsigned num_bins,  char *f, unsigned first, 
	              unsigned last, float *binlimits = 0, float min_v = 0.0, float max_v = 1.0);  
   
	/// Destructor
	~MutualMatrix (); 
	 
	/**
		 It calculates the joint entropy matrix.\\
		 It calculates the  joint entropy matrix. Each element correspond to to the joint entropy between
		 two random variables. It is a simmetric matrix and so is calculated only the inferior triangular
		 matrix without the diagonal.
	*/
  void Calc_JointEntropy_Matrix (void); 

	/**
		 It sums the values of the elements of the joint entropy matrix.\\
		 @return value of the sum
	*/
  float SumElements_JointEntropy_Matrix (void); 
	
	/**
		It prints to stdout the joint entropy matrix.\\
		@see Calc_JointEntropy_Matrix
	*/
  void Print_JointEntropy_Matrix (void); 
	
	/**
		 It calculates the mutual information matrix.\\
		 It calculates the mutual information matrix. Each element correspond to to the mutual
		 information between two random variables. It is a simmetric matrix and so is calculated 
		 only the inferior triangular matrix without the diagonal.
	*/
  void Calc_MutualInfo_Matrix (void);
	
	/**
		 It sums the values of the elements of the mutual information matrix.\\
		 @return value of the sum
	*/
  float SumElements_MutualInfo_Matrix (void); 
	
	/**
		It prints to stdout the mutual information matrix.\\
		@see Calc_MutualInfo_Matrix
	*/
  void Print_MutualInfo_Matrix (void); 
	
	
	/**
		 It calculates the mutual information error matrix without considering single errors.\\
		 It calculates the mutual information error matrix. Each element correspond to to the mutual
		 information error between two random variables not considering single errors. 
		 It is a simmetric matrix and so is calculated 
		 only the inferior triangular matrix without the diagonal.
	*/
  void Calc_MutualInfoError2_Matrix (void); 
	
	/**
		 It sums the values of the elements of the mutual information error matrix 
		 without considering single errors.\\
		 @return value of the sum
	*/
  float SumElements_MutualInfoError2_Matrix (void); 
	
	/**
		It prints to stdout the mutual error information matrix.\\
		@see Calc_MutualInfoError2_Matrix
	*/
  void Print_MutualInfoError2_Matrix (void);
	
	/**
		 It calculates the mutual information error matrix.\\
		 It calculates the mutual information error matrix. Each element correspond to to the mutual
		 information error between two random variables. It is a simmetric matrix and so is calculated 
		 only the inferior triangular matrix without the diagonal.
	*/
  void Calc_MutualInfoError_Matrix (void); 
	
	/**
		 It sums the values of the elements of the mutual information error matrix.\\
		 @return value of the sum
	*/
  float SumElements_MutualInfoError_Matrix (void); 
	
	/**
		It prints to stdout the mutual error information matrix.\\
		@see Calc_MutualInfoError_Matrix
	*/
  void Print_MutualInfoError_Matrix (void); 
	
	/// It returns the current value of the sum of the values of the elements of the mutual information matrix.
	float Read_SumMiM(void) {return SumMiM;}
	
	/// It returns the current value of the sum of the values of the elements of the joint entropy matrix.
	float Read_SumEnM(void) {return SumEnM;}
	
	/// It returns the current value of the sum of the values of the elements of the MeM2 matrix.
	float Read_SumMeM2(void) {return SumMeM2;}
	
	/// It returns the current value of the sum of the values of the elements of the MeM matrix.
	float Read_SumMeM(void) {return SumMeM;}

};

 
#endif
