//  "rssvm_cv.cc"
/*
  Application for the k-fold cross validation of a random subspace ensemble of SVM. The base learners are homogeneous SVM. 
	They can be linear, gaussian, polynomial or sigmoid. 
	In output is printed the error rate, the confusion matrix, the sensitivity, specificity and precision.
	Error rate on each separated fold is reported. Optionally results on the k training sets can be reported.
	Its first argument is datafile. The program expects that datafile.k1.train, datafile.k1.test ...
	datafile.kk.train, datafile.kk.test training and test sets have been previuosly prepared using 
	make_rs_data and do_fold
	Output files:
	- predtest1, ..., predtestk: labels of the predicted classes in the k folds
	- predtest1.ens_errors, ..., predtestk.ens_errors: test error at each iteration on the k folds
	- predtest.average_ens_errors: average test error at each iteration on the k folds
	- predtrain1.ens_errors, ..., predtraink.ens_errors: train error at each iteration on the k folds
	- predtrain.average_ens_errors: average train error at each iteration on the k folds
	- predtest1.blerr, ...,  predtestk.blerr: test error of each base learner on the k folds
	- predtrain1.blerr, ...,  predtraink.blerr: train error of each base learner on the k folds
*/

#include<stdlib.h>
#include<time.h>
#include <sys/times.h>
#include"utils.h"
#include"basesvml.h"
#include"randomsubspace.h"
#include"matrix.h"
#include"vect.h"
#include <unistd.h>
#include <deque.h>

static const char* const BAR = "********************";
static unsigned const SIZE = 256;
static unsigned const DEF_BASE_LEARNERS = 100;
static unsigned const DEF_SUBSPACE_DIM = 2;
static float const DEF_C = 1.0;
static unsigned const DEF_DEGREE = 2;
static float const DEF_S = 1.0;
static float const DEF_R = 1.0;
static float const DEF_SIGMA = 1.0;
static unsigned const DEF_K = 10;
static const char* const BASE_MODEL_NAME = "model";
static const char* const PREDICTED_TRAIN = "predtrain";
static const char* const PREDICTED_TEST = "predtest";
static unsigned type = 0; // SVM type
static unsigned n_base_learners = DEF_BASE_LEARNERS;
static unsigned subspace_dim = DEF_SUBSPACE_DIM;
static float C = DEF_C;
static unsigned degree = DEF_DEGREE;
static float s = DEF_S;
static float r = DEF_R;
static float sigma = DEF_SIGMA;
static unsigned K = DEF_K; // number of folds
static bool b_printout = true;
static bool b_save_base_test_err = false; 
static bool b_save_base_train_err = false; 
static bool b_ens_train_err = false;
static bool b_ens_test_err = false;
static bool b_test_on_train = false;

static char datafile[SIZE];  
static char trainfile[SIZE];
static char testfile[SIZE];
static char basemodel[SIZE]; 
static char predicted_train[SIZE]; 
static char predicted_test[SIZE]; 



///////////////////////////////////  
static char outfile[SIZE];   
static char saverr[SIZE];   

static unsigned savenet = 0;   
static unsigned saveout = 0;   
static int seed = 0;  

static void print_info_use(void);

static int doarg(char* arg[], int i);

////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
	clock_t start, end;
	// time_t start, end, endtrain;
	// struct tms buf1, buf2;
	float err_train, err_test;
	BaseSVM_light_common* base = NULL;
  	
  if (argc < 2) 
     print_info_use();
  if ((argc%2) == 1)
     print_info_use();
		 
	strcpy(datafile, argv[1]);
	strcpy(basemodel, BASE_MODEL_NAME);
	strcpy(predicted_train, PREDICTED_TRAIN);
	strcpy(predicted_test, PREDICTED_TEST);
	
	for (int i=2; i<argc; i+=2) 
     if (doarg(argv, i))
        print_info_use();
	
  // print general info to stdout
	cout << "Cross validation of a random subspace SVM ensemble" << endl;
	cout << "Number of base learners: " << n_base_learners << endl;
	cout << "Subspace dimension: " << subspace_dim << endl;
  cout << "Number of classes: 2" << endl;
	cout << "Number of folds: " << K << endl;
	cout << "Base SVM type : ";
	switch (type) {
	  case 0: cout << "Linear" << endl;
		        cout << "Parameter C = " << C << endl; 
		        base = new BaseSVM_light_linear (C);
		        break;
		case 1: cout << "Polynomial" << endl;
		        cout << "Parameters:" << endl;
						cout << "  degree = " << degree << endl;
						cout << "  s = " << s << endl;
						cout << "  r = " << r << endl;						
						cout << "  C = " << C << endl; 
						base = new BaseSVM_light_poly (degree, C, s, r);
						break;
		case 2: cout << "RBF (gaussian)" << endl;
		        cout << "Parameters:" << endl;
						cout << "  sigma = " << sigma << endl;
						cout << "  C = " << C << endl; 
						base = new BaseSVM_light_gaussian (1.0/sigma, C);
						break;
		case 3: cout << "Sigmoid" << endl;
		        cout << "Parameters:" << endl;
						cout << "  s = " << s << endl;
						cout << "  r = " << r << endl;						
						cout << "  C = " << C << endl;
						base = new BaseSVM_light_sigmoid (C, s, r);
						break;
    default:  cout << "Not known SVM kernel type" << endl;
	            exit(1); 
	}
  	
	if (base == NULL)
	  error_exit(argv[0],  " : Error creating base SVM");
	
	//vectors storing the error on the folds
	vect fold_train_error(K);
	vect fold_test_error(K);
	// global confusion matrix storing the predictions on all the folds
	matrix global_confusion(2,2);
	// list of the error averaged on the k folds at each iteration
	deque<float> avg_train_ens_error(n_base_learners);
	deque<float> avg_test_ens_error(n_base_learners);
	// name of the files storing the preicted classes at each fold
	char pfold_test[SIZE];
	char pfold_train[SIZE];
	
	float avg_train_error = 0;
	float avg_test_error = 0;
	float stdev_train = 0;
	float stdev_test = 0;
	
	float sensitivity = 0;
	float specificity = 0;
	float precision = 0;
	
	RandomSubspace * rsSVM;
	ostrstream otrain(trainfile,256,ios::out);
	ostrstream otest(testfile,256,ios::out);	
	ostrstream opfold_test(pfold_test,256,ios::out);
	ostrstream opfold_train(pfold_train,256,ios::out);

	// starting cross validation
	start = clock();
	for (unsigned k=1; k<=K; k++) {
    // building the bagged ensemble of SVM
	  rsSVM = new RandomSubspace(base);
	  otrain.seekp(0,ios::beg);
		otrain << datafile << ".f" << k << ".train" << ends;
		opfold_train.seekp(0,ios::beg);
		opfold_train << predicted_train <<  k  << ends;
	  // training of the RS SVM ensemble
	  rsSVM->train_base_model (trainfile, basemodel, 1, n_base_learners, false, b_printout);
	  
	
	  // test on the train set if and only if b_test_on_train == true
	  if (b_test_on_train) { 
	    if (b_ens_train_err)   { // save the error of the ensemble at each step    
	      fold_train_error[k-1] = rsSVM->combine_models_step(trainfile, pfold_train, 1, n_base_learners, b_save_base_train_err, b_printout);
				for (int q=0; q<n_base_learners; q++)
		       avg_train_ens_error[q] += rsSVM->ensemble_errors[q];  
			}                                                                                      	          
	    else                                                                                                             
	      fold_train_error[k-1] = rsSVM->combine_models(trainfile, pfold_train, 1, n_base_learners, b_save_base_train_err, b_printout);  		
	  rsSVM->ensemble_errors.clear();
	  }
	
	  // testing the test set
		otest.seekp(0,ios::beg);
		otest << datafile << ".f" << k << ".test" << ends;
		opfold_test.seekp(0,ios::beg);
		opfold_test << predicted_test <<  k  << ends;
	  if (b_ens_test_err)  { // save the error of the ensemble at each step
	     fold_test_error[k-1] = rsSVM->combine_models_step(testfile, pfold_test, 1, n_base_learners, b_save_base_test_err, b_printout, false);        
	     for (int q=0; q<n_base_learners; q++)
		     avg_test_ens_error[q] += rsSVM->ensemble_errors[q];
			}
			else
	      fold_test_error[k-1] = rsSVM->combine_models(testfile, pfold_test, 1, n_base_learners, b_save_base_test_err, b_printout);
		// updating confusion matrix
		global_confusion += rsSVM->get_conf_matrix();		
	  delete rsSVM;
	} // end cross validation
	end = clock();
	
	// computing the average error a each iteration w.r.t. to the K folds
	for (int q=0; q<n_base_learners; q++) {
		 avg_test_ens_error[q] /= K;
		 avg_train_ens_error[q] /= K;
	}
	
	
	// computing the average error  w.r.t. the K folds
	for (unsigned k=0; k<K; k++) {
	  avg_train_error += fold_train_error[k];
		avg_test_error += fold_test_error[k];	
	}
	avg_train_error /= K;
	avg_test_error /= K;
	// computing the standard deviation w.r.t. the K folds
	float x = 0;
	for (unsigned k=0; k<K; k++) {
	  x = avg_train_error - fold_train_error[k];
		stdev_train += (x*x);
		x = avg_test_error - fold_test_error[k];
		stdev_test += (x*x);	
	}
	stdev_train = sqrt(stdev_train/(K-1));
	stdev_test = sqrt(stdev_test/(K-1));
	
	// computing sensitivity etc.
	float TP, TN, FP, FN;
	TP = global_confusion(0,0);
	TN = global_confusion(1,1);
	FP = global_confusion(0,1);
	FN = global_confusion(1,0);
	avg_test_error = 1.0 -((TP+TN)/(TP+TN+FP+FN));
	sensitivity = TP/(TP+FN);
	specificity = TN/(TN+FP);
	precision = TP/(TP+FP);
	
	// writing error at each iteration w.r.t. to the K training file
	if (b_test_on_train && b_ens_train_err) {
	  opfold_test.seekp(0,ios::beg);
		opfold_test << predicted_train <<  ".average_ens_errors"  << ends;
		ofstream of(pfold_test);
	  if (!of)
	    error_exit("rssvm_cv : Can't open file ", pfold_test);
    for (unsigned i = 0; i < avg_train_ens_error.size(); i++)
	    of << avg_train_ens_error[i] << endl;
	  if (of.bad())
	    error_exit("rssvm_cv : Error writing file ", pfold_test);
	}
	// writing error at each iteration w.r.t. to the K test files
	if (b_ens_test_err) {
	  opfold_test.seekp(0,ios::beg);
		opfold_test << predicted_test <<  ".average_ens_errors"  << ends;
		ofstream of(pfold_test);
	  if (!of)
	    error_exit("rssvm_cv : Can't open file ", pfold_test);
    for (unsigned i = 0; i < avg_test_ens_error.size(); i++)
	    of << avg_test_ens_error[i] << endl;
	  if (of.bad())
	    error_exit("rssvm_cv : Error writing file ", pfold_test);
	}
	
	
	// printing results
  print_elapsed_cpu_time(start, end);	
	
	// Results on the k folds:
	// Results on the training sets
	if (b_test_on_train) {
	  cout << BAR << endl;
		cout.precision(4);
	  cout << "Training results on the " << K << " folds:" << endl;
	  for (unsigned k=0; k<K; k++) 
	    cout << "Fold n. " << k+1 << " : " << fold_train_error[k] << endl;
		cout << "Average error : " << avg_train_error << "  Stdev : " << stdev_train << endl;
	}
	
	// Results on the test sets
	cout << endl << BAR << endl;
  cout.precision(4);
	cout << "Testing results on the " << K << " folds:" << endl;
	for (unsigned k=0; k<K; k++) 
	  cout << "Fold n. " << k+1 << " : " << fold_test_error[k] << endl;
  cout << "Average error : " << avg_test_error << "  Stdev : " << stdev_test << endl;
	
	cout << endl << "Confusion matrix:" << endl;
	global_confusion.print(cout,1);
	// Printing sensitivity, specificity, precision:
	cout.width(16); cout << "Sensitivity";
	cout.width(16); cout << "Specificity"; 
	cout.width(16); cout << "Precision" << endl;   
	cout.width(16); cout << sensitivity;
	cout.width(16); cout << specificity; 
	cout.width(16); cout << precision << endl;     
	cout << BAR << endl;
	
	// print summary results on the last row:
  cout.setf(ios::left, ios::adjustfield);
  cout << endl << "Summary results:" << endl;
  cout.width(16);  cout << "Test error";
  cout.width(16);  cout << "St. dev.";
  cout.width(16);  cout << "Train error";
  cout.width(16);  cout << "St. dev.";
  cout.width(16);  cout << "Sensitivity";
  cout.width(16);  cout << "Specificity"; 
  cout.width(16);  cout << "Precision" << endl;  
	cout.width(16);  cout << avg_test_error;
	cout.width(16);  cout << stdev_test;
	cout.width(16);  cout << avg_train_error;
	cout.width(16);  cout << stdev_train;
  cout.width(16);  cout << sensitivity;
  cout.width(16);  cout << specificity;
  cout.width(16);  cout << precision << endl;
	
  delete base;
  return 0;
}



void print_info_use(void)
{
  cerr << "
  rssvm_cv. NEURObjects 2.0 application for the k fold cross validation
            of a random subspace ensemble of SVMs.
  Usage: rssvm_cv datafile [options]
  Options:
  -type unsigned      SVM base learner type:
                      0: linear (default)
                      1: polynomial
                      2: RBF (gaussian)
                      3: sigmoid (MLP-like)
  -nb unsigned        number of base learners  (def. 100)
  -subdim unsigned    dimension of the subspace (def. 2)
  -k unsigned         number of folds (def. 10)
  -c float            regularization parameter of the base SVM (def. 1)
  -deg unsigned       degree of the polynomial SVM (s a*b + r)^d (def. 2)
  -s float            s parameter of the poly SVM or of the sigmoid SVM (def. 1)
  -r float            s parameter of the poly SVM or of the sigmoid SVM (def. 1)
  -sigma float        sigma parameter of the gaussian SVM: exp(-(a-b)^2/sigma)
  -sbmodel string     save the learned base models onto sbmodel1..sbmodeln	
  -sbtest  unsigned   save testing errors of the base learners: 0 (def) no, 1 yes
  -sbtrain  unsigned  save training errors of the base learners: 0 (def) no, 1 yes
  -setrain unsigned   save ensemble train errors at each step: 0 (def) no, 1 yes
  -setest  unsigned   save ensemble test errors at each step: 0 (def) no, 1 yes
  -pred  string       base name for the files saving the classes predicted by the ensemble 
                      with respect to the test and base name for saving ensemble test
                      error at each iteration (def. predtest)
  -predtrain  string  base name for saving ensemble train error at each iteration
                      (def. predtrain)
  -train unsigned     if 0 (def.) testing is not performed on train set, if 1 yes
  -p integer          1 = printing a point at each iteration; 0 = noprint (def.)
  " << endl;
  exit(1);
}


int doarg(char* arg[], int i)
{
  int j;
	if (!strcmp(arg[i], "-type"))
    type = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-nb")) 
    n_base_learners = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-subdim")) 
    subspace_dim = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-k")) 
    K = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-c")) 
    C = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-deg")) 
    degree = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-s")) 
    s = atof(arg[i+1]);
	else if (!strcmp(arg[i], "-r")) 
    r = atof(arg[i+1]);
	else if (!strcmp(arg[i], "-sigma")) {
    sigma = atof(arg[i+1]);
		if (sigma == 0) sigma = 1.0; }
	else if (!strcmp(arg[i], "-sbmodel")) 
    strcpy(basemodel, arg[i+1]);
	else if (!strcmp(arg[i], "-sbtest")) {
    if (atoi(arg[i+1]) == 1) b_save_base_test_err = true; }
  else if (!strcmp(arg[i], "-sbtrain")) {
    if (atoi(arg[i+1]) == 1) b_save_base_train_err = true; }
	else if (!strcmp(arg[i], "-setrain")) {
    if (atoi(arg[i+1]) == 1) b_ens_train_err = true; }
	else if (!strcmp(arg[i], "-setest")) {
    if (atoi(arg[i+1]) == 1) b_ens_test_err = true;} 
  else if (!strcmp(arg[i], "-pred"))  
		strcpy(predicted_test, arg[i+1]);
  else if (!strcmp(arg[i], "-predtrain"))  
    strcpy(predicted_train, arg[i+1]);
	else if (!strcmp(arg[i], "-train")) {
    if (atoi(arg[i+1]) == 1) b_test_on_train = true; }
	else if (!strcmp(arg[i], "-p")) {
    j = atoi(arg[i+1]);
		if (j == 0)
		   b_printout = false;}
  else 
    return 1;
  return 0;
} 
