// stattest.cc

#include"stattest.h"
#include"utils.h"

McNemar::McNemar (char* p1, char* p2, char* t) {

 load_labels(p1, pred1);
 load_labels(p2, pred2);
 load_labels(t, target);
 if ((pred1.size() != pred2.size()) || (pred2.size() != target.size()))
   error_exit("Error in prediction files");
 n01 = n10 = 0; 
 build_cont_table();  
 gamma = compute_pearson();
}


void McNemar::load_labels(char* file, deque<int>& pred) {
  
  int a;
  ifstream ifs(file);
  if (!ifs)
    error_exit("Error opening ", file);
    
  ifs >> a;
  while (!ifs.eof()) {
    pred.push_back(a);
    ifs >> a;
  }  
  if (ifs.bad())
    error_exit("Error reading ", file);
}

/// It builds the contingency table
void McNemar::build_cont_table(void) {
  int a,b;
  int n = pred1.size();
  cont_tab[0][0] = cont_tab[0][1] = cont_tab[1][0] = cont_tab[1][1] = 0;
  for (int i=0; i<n; i++) {
    if (pred1[i] == target[i])
      a=1;
    else
      a=0;
    if (pred2[i] == target[i])
      b=1;
    else
      b=0;
    cont_tab[a][b]++;
  }
  n01 = cont_tab[0][1];
  n10 = cont_tab[1][0];
}


/**
   Computes the  Pearson statistic of the Mc Nemar's test.\\
   @return value of the computed Pearson statistic
*/
float McNemar::compute_pearson(void) {
  
  float res;
  float x; 
  if (n01 == n10)
    return 0; 
  x = abs(n01 - n10) - 1;
  res = (x*x)/(float)(n01+n10);
  return res;
}


/**
   Test of the null hypothesis.\\
   The null hypothesis (that is the two classifiers are equivalent) is rejected 
   if the result of the Pearson statistic is larger than value of the chi square function with 1 degree
   of freedom and level of confidence 0.1.
   The actual level of confidence is returned in level. Level is set to 1 if the null hypothesis 
   cannot be rejected.
   @param level level of confidence
   @return true if the null hypothesis is rejected, otherwise false
*/
bool McNemar::Reject_null(float& level) {
  
  level = 1.0;
  bool res;
  //if (degree != beta95)
  //  error_exit ("Reject_null: degree not supported");
  //if (gamma > chi95)
  //  res = true;
  //else
  //  res = false;
  //return res;
  for (int i = 0; i < nValues; i++) {
    if (gamma >= chiValues[i]) {
      level = pValues[i];
      break;
    }
  }
  if (level < 1.0)
    res = true;
  else
    res = false;
  return res;
}



///////////////////////////////////////////////////////////////////////////////

ResPairT::ResPairT(char* f1, char* f2) {
 t_stat = p_avg = p_sd = 0;
 load_errors(f1, p1);
 load_errors(f2, p2);
 if (p1.size() != p2.size())
   error_exit("ResPairT : ", "Number of trials do not agree.");
 n = p1.size(); 
}


void ResPairT::load_errors(char* file, deque<float>& misrate) {
  
  float a;
  ifstream ifs(file);
  if (!ifs)
    error_exit("Error opening ", file);
    
  ifs >> a;
  while (!ifs.eof()) {
    misrate.push_back(a);
    ifs >> a;
  }  
  if (ifs.bad())
    error_exit("Error reading ", file);
}


void ResPairT::calc_p(void){  
	float x;
  for (int i=0; i<n; i++) {
	  x = p1[i] - p2[i];
		p.push_back(x);
	}
}

float ResPairT::calc_avg(void){
  float m = 0;
	for (int i=0; i<n; i++) 
    m += p[i];
	m /= n;
	return (p_avg = m);
}

float ResPairT::calc_sd(void){
  float x;
	float sd = 0;
  for (int i=0; i<n; i++) {
	  x = p_avg - p[i];
		sd += (x*x);
	}
	return (p_sd = sqrt(sd/(n-1)));
}


float ResPairT::calc_t_stat(void){
	calc_p();
	p_avg = calc_avg();
	p_sd = calc_sd();
	t_stat = fabs((p_avg * sqrt((float)n)) / p_sd);
	return t_stat;
}


bool ResPairT::Reject_null(int conf) {
  
  bool res;
  calc_t_stat();
  if (t_stat > t[n-1][conf])
    res = true;
  else
    res = false;
  return res;
}


float ResPairT::test(void) {
  float res;	
  calc_t_stat();
	int i = c99;
	while (i >= c90) {
	  if (t_stat > t[n-1][i])
		  break;	
		i--;
  }
	switch (i) {
	case c99  : res = 0.99; break;
	case c975 : res = 0.975; break;
	case c95  : res = 0.95; break;
	case c90  : res = 0.90; break;
	default    : res = 0;	
	}
	return res;
}


