//  "svm_cv.cc"
// May 2004
// modified october 2004

/*
  Application for the k-fold cross validation of SVMs. 
	The SVMs can be linear, gaussian, polynomial or sigmoid. 
	In output is printed the error rate, the confusion matrix, the sensitivity, specificity and precision.
	Error rate on each separated fold is reported, both on the training and test sets.
	Output files:
	- predtest1, ..., predtestk: labels of the predicted classes in the k folds
*/

#include<stdlib.h>
#include<time.h>
#include <sys/times.h>
#include <unistd.h>
#include <deque.h>
#include <string>
#include"utils.h"
#include"basesvml.h"
#include"matrix.h"
#include"vect.h"


static const char* const BAR = "********************";
static unsigned const SIZE = 256;
static float const DEF_C = 1.0;
static float const DEF_COST = 1.0;
static unsigned const DEF_DEGREE = 2;
static float const DEF_S = 1.0;
static float const DEF_R = 1.0;
static float const DEF_SIGMA = 1.0;
static unsigned const DEF_K = 10;
static const char* const MODEL_NAME = "model";
static const char* const PREDICTED_TRAIN = "predtrain";
static const char* const PREDICTED_TEST = "predtest";
static unsigned type = 0; // SVM type
static float C = DEF_C;
static float cost = DEF_COST;
static unsigned degree = DEF_DEGREE;
static float s = DEF_S;
static float r = DEF_R;
static float sigma = DEF_SIGMA;
static unsigned K = DEF_K; // number of folds
static bool b_printout = true;

static char datafile[SIZE];  
static char trainfile[SIZE];
static char testfile[SIZE];
static char model[SIZE];
static char predicted_train[SIZE]; 
static char predicted_test[SIZE]; 



///////////////////////////////////  
static void print_info_use(void);

static int doarg(char* arg[], int i);

/* It updates the confusion matrix according to the testfile (in svmlight-format) and the predicted classes
/ stored in pfold_test (in NEURObjects format) */
static void update_confusion_matrix(matrix & global_confusion, char* testfile, char* pfold_test);

////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
	clock_t start, end;
	// time_t start, end, endtrain;
	// struct tms buf1, buf2;
	float err_train, err_test;
	BaseSVM_light_common* svm = NULL;
  	
  if (argc < 2) 
     print_info_use();
  if ((argc%2) == 1)
     print_info_use();
		 
	strcpy(datafile, argv[1]);
	strcpy(model, MODEL_NAME);
	strcpy(predicted_train, PREDICTED_TRAIN);
	strcpy(predicted_test, PREDICTED_TEST);
	
	for (int i=2; i<argc; i+=2) 
     if (doarg(argv, i))
        print_info_use();
	
  // print general info to stdout
	cout << "SVM cross validation." << endl;
  cout << "Number of classes: 2" << endl;
	cout << "Number of folds: " << K << endl;
	cout << "Cost factor: " << cost << endl;
	cout << "Base SVM type : ";
	switch (type) {
	  case 0: cout << "Linear" << endl;
		        cout << "Parameter C = " << C << endl; 
		        svm = new BaseSVM_light_linear (C, cost);
		        break;
		case 1: cout << "Polynomial" << endl;
		        cout << "Parameters:" << endl;
						cout << "  degree = " << degree << endl;
						cout << "  s = " << s << endl;
						cout << "  r = " << r << endl;						
						cout << "  C = " << C << endl; 
						svm = new BaseSVM_light_poly (degree, C, s, r, cost);
						break;
		case 2: cout << "RBF (gaussian)" << endl;
		        cout << "Parameters:" << endl;
						cout << "  sigma = " << sigma << endl;
						cout << "  C = " << C << endl; 
						svm = new BaseSVM_light_gaussian (1.0/sigma, C, cost);
						break;
		case 3: cout << "Sigmoid" << endl;
		        cout << "Parameters:" << endl;
						cout << "  s = " << s << endl;
						cout << "  r = " << r << endl;						
						cout << "  C = " << C << endl;
				  	svm = new BaseSVM_light_sigmoid (C, s, r, cost);
						break;
    default:  cout << "Not known SVM kernel type" << endl;
	            exit(1); 
	}
  	
	if (svm == NULL)
	  error_exit(argv[0],  " : Error creating SVM");
	
	//vectors storing the error on the folds
	vect fold_train_error(K);
	vect fold_test_error(K);
	// global confusion matrix storing the predictions on all the folds
	matrix global_confusion(2,2);
	
	// name of the files storing the preicted classes at each fold
	char pfold_test[SIZE];
	
	float avg_train_error = 0;
	float avg_test_error = 0;
	float stdev_train = 0;
	float stdev_test = 0;
	
	float sensitivity = 0;
	float specificity = 0;
	float precision = 0;
	
	
	ostrstream otrain(trainfile,256,ios::out);
	ostrstream otest(testfile,256,ios::out);	
	ostrstream opfold_test(pfold_test,256,ios::out);

	// starting cross validation
	start = clock();
	for (unsigned k=1; k<=K; k++) {
    // preparing the name of the training files
	  otrain.seekp(0,ios::beg);
		otrain << datafile << ".f" << k << ".train" << ends;
	
	  // training of the SVM 
	  fold_train_error[k-1] = svm->train_learner (trainfile, model);
	
	  // preparing the names of the test files and of the predicted classes files.
		otest.seekp(0,ios::beg);
		otest << datafile << ".f" << k << ".test" << ends;
		opfold_test.seekp(0,ios::beg);
		opfold_test << predicted_test <<  k  << ends;
	  
		// testing the test set
	  fold_test_error[k-1] = svm->test_learner(testfile, model, pfold_test);
		// updating confusion matrix
	  update_confusion_matrix(global_confusion, testfile, pfold_test);
		if (b_printout)
		  cout << ".";
	} // end cross validation
	end = clock();
	cout << endl;
	
	// computing the average error  w.r.t. the K folds
	for (unsigned k=0; k<K; k++) {
	  avg_train_error += fold_train_error[k];
		avg_test_error += fold_test_error[k];	
	}
	avg_train_error /= K;
	avg_test_error /= K;
	// computing the standard deviation w.r.t. the K folds
	float x = 0;
	for (unsigned k=0; k<K; k++) {
	  x = avg_train_error - fold_train_error[k];
		stdev_train += (x*x);
		x = avg_test_error - fold_test_error[k];
		stdev_test += (x*x);	
	}
	stdev_train = sqrt(stdev_train/(K-1));
	stdev_test = sqrt(stdev_test/(K-1));
	
	// computing sensitivity etc.
	float TP, TN, FP, FN;
	TP = global_confusion(0,0);
	TN = global_confusion(1,1);
	FP = global_confusion(0,1);
	FN = global_confusion(1,0);
	avg_test_error = 1.0 -((TP+TN)/(TP+TN+FP+FN));
	sensitivity = TP/(TP+FN);
	specificity = TN/(TN+FP);
	precision = TP/(TP+FP);
	
	// printing results
  print_elapsed_cpu_time(start, end);	
	
	// Results on the k folds:
	// Results on the training sets

	cout << BAR << endl;
  cout.precision(4);
	cout << "Training results on the " << K << " folds:" << endl;
	for (unsigned k=0; k<K; k++) 
	  cout << "Fold n. " << k+1 << " : " << fold_train_error[k] << endl;
	cout << "Average error : " << avg_train_error << "  Stdev : " << stdev_train << endl;
	
	// Results on the test sets
	cout << endl << BAR << endl;
  cout.precision(4);
	cout << "Testing results on the " << K << " folds:" << endl;
	for (unsigned k=0; k<K; k++) 
	  cout << "Fold n. " << k+1 << " : " << fold_test_error[k] << endl;
  cout << "Average error : " << avg_test_error << "  Stdev : " << stdev_test << endl;
	
	cout << endl << "Confusion matrix:" << endl;
	global_confusion.print(cout,1);
	// Printing sensitivity, specificity, precision:
	cout.width(16); cout << "Sensitivity";
	cout.width(16); cout << "Specificity"; 
	cout.width(16); cout << "Precision" << endl;   
	cout.width(16); cout << sensitivity;
	cout.width(16); cout << specificity; 
	cout.width(16); cout << precision << endl;     
	cout << BAR << endl;
	
  
	// print summary results on the last row:
  cout.setf(ios::left, ios::adjustfield);
  cout << endl << "Summary results:" << endl;
  cout.width(16);  cout << "Test error";
  cout.width(16);  cout << "St. dev.";
  cout.width(16);  cout << "Train error";
  cout.width(16);  cout << "St. dev.";
  cout.width(16);  cout << "Sensitivity";
  cout.width(16);  cout << "Specificity"; 
  cout.width(16);  cout << "Precision" << endl;  
	cout.width(16);  cout << avg_test_error;
	cout.width(16);  cout << stdev_test;
	cout.width(16);  cout << avg_train_error;
	cout.width(16);  cout << stdev_train;
  cout.width(16);  cout << sensitivity;
  cout.width(16);  cout << specificity;
  cout.width(16);  cout << precision << endl;
	
  delete svm;
  return 0;
}



void update_confusion_matrix(matrix & global_confusion, char* testfile, char* pfold_test) {
  
	int trueclass, predclass;
	
	ifstream ftest(testfile);
  if (!ftest) error_exit("svm_cv: Cannot open test file ", testfile);
	ifstream ffold(pfold_test);
  if (!ffold) error_exit("svm_cv: Cannot open predicted classes file", pfold_test);

  // read input files and update confusion matrix
	string line;
	getline(ftest,line); getline(ftest,line); 
	ftest >> trueclass; 
	getline(ftest,line);
	ffold >> predclass;
	while (!ftest.eof() && !ffold.eof()){	  
	  if (trueclass !=1)
	    trueclass = 2;	  
	  global_confusion(predclass-1, trueclass-1)+=1;
		ftest >> trueclass; 
	  getline(ftest,line);
	  ffold >> predclass;
	} 
	
  if (ftest.bad())
    error_exit("Error reading test file ", testfile);
	if (ffold.bad())
    error_exit("Error reading predicted classes file", pfold_test);
}



void print_info_use(void)
{
  cerr << "
  svm_cv. NEURObjects 2.0 application for the SVM k fold cross validation.
  Usage: svm_cv datafile [options]
  Options:
  -type unsigned      kernel type:
                      0: linear (default)
                      1: polynomial
                      2: RBF (gaussian)
                      3: sigmoid (MLP-like)
  -k unsigned         number of folds (def. 10)
  -cost float         cost-factor, for asymmetric cost error (def. 1.0)
  -c float            regularization parameter of the base SVM (def. 1)
  -deg unsigned       degree of the polynomial SVM (s a*b + r)^d (def. 2)
  -s float            s parameter of the poly SVM or of the sigmoid SVM (def. 1)
  -r float            r parameter of the poly SVM or of the sigmoid SVM (def. 1)
  -sigma float        sigma parameter of the gaussian SVM: exp(-(a-b)^2/sigma)
  -pred  string       base name for the files saving the classes predicted by the SVM 
                      with respect to each fold (test set) 
  -predtrain  string  base name for the files saving the classes predicted by the SVM 
                      with respect to each fold (train set) 
  -p integer          1 = printing a point at each iteration; 0 = noprint (def.)
  " << endl;
  exit(1);
}


int doarg(char* arg[], int i)
{
  int j;
	if (!strcmp(arg[i], "-type"))
    type = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-k")) 
    K = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-cost")) 
    cost = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-c")) 
    C = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-deg")) 
    degree = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-s")) 
    s = atof(arg[i+1]);
	else if (!strcmp(arg[i], "-r")) 
    r = atof(arg[i+1]);
	else if (!strcmp(arg[i], "-sigma")) {
    sigma = atof(arg[i+1]);
		if (sigma == 0) sigma = 1.0; }
  else if (!strcmp(arg[i], "-pred"))  
		strcpy(predicted_test, arg[i+1]);
  else if (!strcmp(arg[i], "-predtrain"))  
    strcpy(predicted_train, arg[i+1]);
	else if (!strcmp(arg[i], "-p")) {
    j = atoi(arg[i+1]);
		if (j == 0)
		   b_printout = false;}
  else 
    return 1;
  return 0;
} 
