#ifndef TSET_H
#define TSET_H

#include<iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <string.h>
#include <ctype.h>
#include <vector.h>
#include <deque.h>
#include <map.h>
#include "matrix.h"

 // Suffix for the training set file
	 static const char* const FOLD_TRAIN_SUFFIX = ".train";
 // Suffix for the test set file
	 static const char* const FOLD_TEST_SUFFIX = ".test";

/**
   Class for reading and loading the training set from file.\\
   Attributes are loaded in the matrix data in numerical form.
	 Classes are loaded in memory in numerical form in the vector target.
	 Each line of the file to be read must be a sample, and each different attribute of the same
	 sample must be separated by a separator or a sequence of blanks. The last field of the line is the class:
   attr1,attr2, ..., attrn_attr,target
*/ 
class TrainingSet {

public:
   
	 /// position order of the class in the rows of the data file.
   enum class_pos {first, last, noclass}; 
	 
	 /// format type of the data: light--> SVMlight; NO --> NEURObjects
   enum typeformat {light, NO}; 
	 
	
	 
	 /** Blank separator. 
	 If separator is BLANK, attributes and classes can be separated by an arbitrary
	 sequence of blanks
	 */
	 static const char BLANK = ' ';
   
	 /// Dummy constructor
   TrainingSet (void)
	 {n_attr = n_train = n_class = is_normalized = 0;}
	 
	 /**
	 Constructor.
	 @param num_attr attribute cardinality
	 @param num_train training set cardinality
	 @param separator attribute separator in the rows of the input data file
	 @param pos position order of the class in the rows of the data file
	 */
   TrainingSet ( unsigned num_attr, unsigned num_train, char separator = ',', class_pos pos = last);
	 
	 /**
	 Constructor.
	 It reads from file the attributes and the class of the patterns.
	 If separator is BLANK, the attributes and the class can be separated by an arbitrary sequence of blanks,
	 otherwise by the selected separator character.
	 @param num_attr attribute cardinality
	 @param num_train training set cardinality
	 @param file name of the input data file
	 @param separator attribute separator in the rows of the input data file
	 @param pos position order of the class in the rows of the data file
	 */
   TrainingSet ( unsigned num_attr, unsigned num_train,  char *file, char separator = ',',
	               class_pos pos = last) ;
	 /// Destructor
   virtual ~TrainingSet () ;
	 
	 /**
   It normalizes input data.\\
   Input data are normalized in such a way that the mean value is zero and standard deviation 
   sv is 1: \\
   normalized_x = (x - mean(x)) / sv.
   It computes mean and standard deviation of the training set values
   */
   void normalize(void); 
   
    /**
   It normalizes input data.\\
   Input data are normalized in such a way that the mean value is zero and the variance is 1. 
   */
   void normalize_var(void); 
   
	 /**
     It normalizes input data.\\
		 Input data are normalized in such a way that the mean value is zero and standard deviation 
		 sv is 1: \\
     normalized_x = (x - mean(x)) / sv.
		 In order to perform the normalization it uses the given vectors of the mean and standard deviation.
     @param m mean vector of attributes before normalization. 
     @param sv standard deviation vector before normalization.\\  
   */
   void normalize(vect& m, vect& sv); 
   
	 // Computes mean and standard deviation of the data
   void calc_mean_stdev(void) { mean_calc(); stdev_calc();}
	 
	 /**
     It subdivides a file for k-fold cross-validation.\\
     @param numfold number of folds
     @param name base name of the file
   */
   void FoldRand(unsigned numfold, char* name); 
	 
	 /**
     Subsampling with replacement of the data, saving results into a file.\\
		 It saves in the file name a subset of the data drawn with replacement according to a uniform probability distribution
		 @param name  name of the file to save data
     @param numsample number of the samples 
		 @param initseed if equal to 0 the seed of the random generator is initialized using the computer clock, otherwise is set to
		                 initseed
   */
   void Subsample_with_replacement (char* name, unsigned numsample, long initseed = 0); 
   
	 /**
     Subsampling with replacement of the data.\\
		 It generates a subset of the data drawn with replacement according to a uniform probability distribution
     @param numsample number of the samples to be generated
		 @param initseed if equal to 0 the seed of the random generator is initialized using the computer clock, otherwise is set to
		                 initseed
		 @return the subsampled training set
   */
	 TrainingSet* Subsample_with_replacement (unsigned numsample, long initseed = 0);
	 
	 /**
     It stores the data set into a file.\\ 
     @param name file name      
   */
	 void save(char* name);
	 
	 /**
     It stores the data set into a file in SVM-light format.\\ 
		 All the classes with target <> 1 are relabeled with -1.
     @param name file name      
   */
	 void save_light_format(char* name); 
   
   /**
     Set a low threshold and a ceiling to the data.\\ 
		 @param low_thresh lower bound to which data are set
     @param hi_tresh  ceiling of the data   
   */
   void set_thresholds(float low_thresh, float hi_tresh);
	 
   inline unsigned Ntrain() const {return n_train;}
   inline unsigned Nattr() const {return n_attr;}
   inline unsigned Nclass() const {return n_class;}
   inline vect& Read_target() {return target;}
   inline matrix& Read_data() {return data;}
   inline void Set_data(matrix& m) {data = m; n_train = m.dim1(); n_attr = m.dim2();}
   inline vect& Read_mean(void)  {return mean;}
   inline vect& Read_stdev(void) {return stdev;}   
   inline unsigned Is_normalized() const {return is_normalized;}
   inline void print_data(); 
   inline void print_target(); 
   inline void print_info();
   inline void print_sep() {cout << "Separator = " << sep ;}
	 inline char read_sep() {return sep ;}
	 inline void set_class_pos(class_pos pos) { cl_position = pos ;}
	 inline class_pos get_class_pos() { return cl_position;}
	 
	 /**
     Set the data set separator character \\ 
     @param c separator character     
   */ 
	 inline void set_sep(char c) {sep = c ;}
   /**
     Transpose the data matrix \\ 
     It swaps the rows and the columns of the data matrix. 
    */ 
   void Transpose_data(void); 
   /**
     Load the file of the targets \\ 
     @param t name of the target file
    */ 
   void load_target(char* t);
	 
	 /**
     Save targets into a file \\ 
		 The labels are stored one for each line.
     @param t name of the target file
    */ 
   void save_target(char* t);
   
   /**
	  Generate a k-dimensional training set, with k<nattr, randomly keeping values from the original training set.\\
		@param k number of the dimensions to be randomly drawn
    @param v vector storing the indices of the randomly selected dimensions
		@return k-dimensional training set 		
  */
   TrainingSet* generate_random_subspace(int k, unsigned v[]);
	 
	 /**
	  Generate a k-dimensional training set, with k<nattr, using a predefined set of indices.\\
		It gets the indices of the dimensions from the vector v.
		@param k number of the dimensions to be randomly drawn
    @param v vector storing the indices of the randomly selected dimensions
		@return k-dimensional training set 		
  */
   TrainingSet* generate_predefined_random_subspace(int k, unsigned const v[]);
  
	/**
	  Generate m k-dimensional training sets, randomly selecting k dimensions from the original input space.\\ 
		It calls m times the method generate_random_subspace, to generate m training sets with dimension k. The training
	  sets are stored in m file (filename1... filenamem). It generate also the file filename.index containing the values 
		of the randomly selected dimensions.\\
		Format of the file filename.index:
		m k
		i11 i12 ... i1k
		i21 i22 ... i2k
		...
		im1 im2 ... imk
		where iju  1<=j<=m, 1<=u<=k represents the jth randomly selected index of uth file.
		@param filename base name of the files to be generated
		@param m number of the files to be generated
		@param k number of the dimensions to be randomly drawn
    @param tformat format of the file to be generated: NO--> NEURObjects format; light --> SVMlight format	 
  */

   void generate_multiple_random_subspace(char* filename, int m, int k, typeformat tformat);
	 
	 /**
	  Generate m k-dimensional training sets, using the dimensions stored in a file.\\ 
		It generates m training sets with dimension k, using the indices stored in the file filename.index. The data
	  sets are stored in m file (filename1... filenamem).\\
		Format of the file filename.index:
		m k
		i11 i12 ... i1k
		i21 i22 ... i2k
		...
		im1 im2 ... imk
		where iju  1<=j<=m, 1<=u<=k represents the jth randomly selected index of uth file.
		@param filename base name of the files to be generated
    @param tformat format of the file to be generated: NO--> NEURObjects format; light --> SVMlight format	 
  */

   void generate_predefined_multiple_random_subspace(char* filename, typeformat tformat);
	 
	 /**
	  Filtering of the samples according to the maximal absolute deviation between two attributes.\\
		It filters the samples if the maximal absolute difference between the maximum and minimum values of their
		attributes is above a threshold.
		@param abs_tresh absolute threshold
    @param filtered list of the filtered genes. The filtered genes are those that passed the filter
  */
	 void absolute_filter(float abs_tresh, deque <unsigned> & filtered);
	 
	 /**
	  Filtering of the samples according to the maximal fold change between two attributes.\\
		It filters the samples if the maximal fold change between the attributes is above a threshold.
		@param fold_tresh fold change threshold
    @param filtered list of the filtered genes. The filtered genes are those that passed the filter
  */
	 void fold_filter(float fold_tresh, deque <unsigned> & filtered);
	 
	 /**
	  Filtering of the samples according to the maximal absolute deviation and to the maximal fold 
		change between two attributes.\\
		It filters the samples if the maximal absolute difference between the maximum and minimum values of their
		attributes and if the maximal fold change between the attributes is above a threshold.
		@param abs_tresh absolute threshold
		@param fold_tresh fold change threshold
    @param filtered list of the filtered genes. The filtered genes are those that passed the filter
		@return data set with the filtered samples
  */
	 TrainingSet * absolute_fold_filter(float abs_tresh, float fold_tresh, deque <unsigned> & filtered);

protected:
   /// Training set cardinality
   unsigned n_train ; 
	 /// Attributes cardinality
   unsigned n_attr ; 
	 /// Cardinality  of the classes   
   unsigned n_class;      
	 /// character separator between attributes in the rows of the input data file
	 char sep;
   vect target;    // Vettore numerico delle classi target
   matrix data;      // Matrice di double per la memorizzazione dei campioni. 
                     // N.righe = ntrain
                     // N.colonne = n_attr 
   vect mean;      // valor medio degli input (dim: n_attr) 
   vect stdev;     // deviazione standard degli input (dim: n_attr) 
   unsigned is_normalized;  //variabile booleana: =1 input normalizzato; =0 input non normalizzato.
	 /// Position of the class label in the data file
	 class_pos cl_position;
   
	 /** Read a file containing the training set and load input data.\\
	     It loads input attributes data on matrix data and numerical target data identifying the
			 classes in vector target. It returns the number of classes deduced by the maximum
			 number associated to a class in the input file.
			 @param file name of the input data file
			 @param pos position of the target class: it can be at the end (last, default) or at the
			 beginning (first) of each line of the input file.
			 @return number of classes
	 */           
   virtual unsigned read(char *file, class_pos pos = last);   
   
	 /// It computes then mean of the input pattern data                                                          
   void mean_calc(void);
	 
	 /// It computes then standard deviation of the input pattern data    
   void stdev_calc(void);
   
	 /**
     It saves train and validation set (n-fold cross - validation).\\
		 If v[i] = 1, data(i,:) and target t[i] are saved in the 
		 file name.fn.test,
     otherwise they are saved in the file name.fn.train.     
     @param name base name of the couple train and test data file
     @param n nth fold
     @param v vector to assign the patterns to the training or test set 
   */
   void DoTestTrainFile(char* name, unsigned n, vector<unsigned>& v); 
   
   /**
    It saves a single pattern data and its class target k on a stream.\\
    @param fdata stream to save data
    @param patt pattern to be saved
    @param k target class to be saved
   */  
   void save_pattern(ofstream& fdata, vect& patt, unsigned k);
	 
	 /**
    It saves a single pattern data and its class target k on a stream in SVM-light format.\\
    @param fdata stream to save data
    @param patt pattern to be saved
    @param k target class to be saved. It can be only 1 or -1.
   */  
	 void save_pattern_light_format(ofstream& fdata, vect& patt, int k);
	 
	 /**
    It saves a subset of data.\\
		It saves numsample samples, using the indices in data and target stored in subset
    @param name name of the file for the data to be saved
    @param numsample number of the samples to be saved
    @param subset indices of the samples in data and target to be saved
   */  
   void save_subset(char* name, unsigned numsample, vect & subset);
	 
	 /**
    It generates a subset of data.\\
		It gebnerates numsample samples, using the indices in data and target stored in subset
    @param numsample number of the samples to be saved
    @param subset indices of the samples in data and target to be saved
		@return training set generated
   */  
	 TrainingSet* generate_subsample(unsigned numsample, vect & subset);
	 
} ;



/**
   Class for building a data set. \\
	 The class label may be a string\\
   @memo Class for building a data set - 	 HTML documentation under construction.
*/
class TSetStrClass : public TrainingSet {

public:   
   TSetStrClass ( unsigned num_attr, unsigned num_train, char separator = ',');
   TSetStrClass ( unsigned num_attr, unsigned num_train, char *file, 
                  char separator = ',', class_pos pos = last) ;
   TSetStrClass ( unsigned num_attr, unsigned num_train, char *file,
                  map <string,unsigned>& dict, char separator = ',', class_pos pos = last);
   ~TSetStrClass () {};
   void print_dict();  
   inline map <string,unsigned>& read_dict() {return D;}

protected:
   map <string,unsigned> D;  
   
                                                                                                           
   void TranslateStringClass (string s, unsigned& n, int i); 
   
   void AssignStringClass(string s, unsigned& n, int i); 
   
   unsigned read(char *file, unsigned build_dict, class_pos pos = last);   
                                                         
} ;


#endif
